import unittest
import six

import ssbio.databases.kegg
from ssbio.databases.kegg import KEGGProp

class TestKEGG(unittest.TestCase):
    """Unit tests for KEGG
    """

    def test_kegg_prop(self):
        kg = KEGGProp(kegg_id='mtu:Rv0417', sequence_path='test_files/sequences/mtu-Rv0417.faa',
                      metadata_path='test_files/sequences/mtu-Rv0417.kegg')

        kg_keys = ['bigg','gene_name','id','kegg','metadata_file','metadata_path','pdbs','refseq',
                   'sequence_file','seq_len','sequence_path','taxonomy','uniprot']

        for k in kg_keys:
            self.assertTrue(hasattr(kg, k))

        self.assertEqual(kg.bigg, None)
        self.assertEqual(kg.gene_name, None)
        self.assertEqual(kg.id, 'mtu:Rv0417')
        self.assertEqual(kg.kegg, 'mtu:Rv0417')
        self.assertEqual(kg.metadata_file, 'mtu-Rv0417.kegg')
        self.assertEqual(kg.pdbs, None)
        self.assertEqual(kg.refseq, 'NP_214931')
        self.assertEqual(kg.sequence_file, 'mtu-Rv0417.faa')
        self.assertEqual(kg.seq_len, 252)
        self.assertEqual(kg.taxonomy, 'mtu  Mycobacterium tuberculosis H37Rv')
        six.assertCountEqual(self, kg.uniprot, ['P9WG73', 'I6Y3Q4'])


from ssbio import utils
date = utils.Date()

import os
import os.path as op
import pandas as pd
import xmltodict
import requests
import zipfile
import shutil
import collections
from tqdm import tqdm

drugbank_xml = 'http://www.drugbank.ca/releases/4-5-0/downloads/all-full-database'
drugbank_xsd = 'http://www.drugbank.ca/docs/drugbank.xsd'


# TODO: check out https://github.com/dhimmel/drugbank

def download_drugbank_db(username, password, outdir=None):
    """Download the DrugBank XML database and return the unzipped file path

    You need to have a DrugBank account now to download the XML file.
    Args:
        username: your DrugBank.ca email username
        password: your DrugBank.ca password
        outdir: optional output directory (default is the current directory)

    Returns:
        rename_unzipped_file (str): path to unzipped XML file

    """
    # TODO: obfuscate password?

    if not outdir:
        outdir = os.getcwd()

    # define filenames here
    outfile = "{}-drugbank_all_full_database.xml.zip".format(date.short_date)
    unzipped_file = 'full database.xml'
    rename_unzipped_file = "{}-drugbank_all_full_database.xml".format(date.short_date)

    if outdir:
        outfile = op.join(outdir, outfile)
        unzipped_file = op.join(outdir, unzipped_file)
        rename_unzipped_file = op.join(outdir, rename_unzipped_file)

    # if the file was already downloaded today, just return it
    if op.exists(rename_unzipped_file):
        return rename_unzipped_file

    r = requests.get(drugbank_xml, auth=(username, password))

    # an HTML page is returned if there is an error getting the file
    if r.headers['Content-Type'] == 'text/html; charset=utf-8':
        raise ConnectionError('DrugBank execute_dir could not be downloaded. Please check your username and password.')

    with open(outfile, "wb") as db:
        db.write(r.content)

    zip_ref = zipfile.ZipFile(outfile, 'r')
    zip_ref.extractall(outdir)
    zip_ref.close()

    shutil.move(unzipped_file, rename_unzipped_file)

    return rename_unzipped_file


def parse_drugbank_xml(db_xml_path, outdir=None):
    """Parse the main DrugBank XML database

    Args:
        db_xml_path:
        outdir:

    Returns:
        Paths to Pandas DataFrames for:
        db_main_path: Main DrugBank ID mapping
        db_bind_path: DrugBank binding partners
        db_snp_path: Pharmacogenomics information

    TODO:
        - Use ElementTree instead

    """
    if not outdir:
        outdir = os.getcwd()

    db_main_path = op.join(outdir, '{}-db_main_df.csv'.format(date.short_date))
    db_bind_path = op.join(outdir, '{}-db_binding_df.csv'.format(date.short_date))
    db_snp_path = op.join(outdir, '{}-db_snp_df.csv'.format(date.short_date))

    if op.exists(db_main_path) and op.exists(db_bind_path) and op.exists(db_snp_path):
        return db_main_path, db_bind_path, db_snp_path

    db_xml = op.join(db_xml_path)
    db_xml_open = open(db_xml, "rb")
    db_dict = xmltodict.parse(db_xml_open)

    db_main = []
    db_bind_to = []
    db_snp = []

    for drug in tqdm(db_dict['drugbank']['drug']):  # [:100]:
        main_row = collections.OrderedDict()

        db_id = utils.force_list(drug['drugbank-id'])[0]['#text']

        main_row['db_id'] = db_id
        main_row['db_name'] = drug['name']
        main_row['db_type'] = drug['@type']
        main_row['db_groups'] = utils.force_string(drug['groups']['group'])

        # Percentage of the drug that is bound in plasma proteins
        main_row['db_protein_binding'] = drug['protein-binding']

        # PharmGKB ID
        if drug['external-identifiers']:
            for ext_id in utils.force_list(drug['external-identifiers']['external-identifier']):
                if ext_id['resource'] == 'PharmGKB':
                    main_row['db_pgkb'] = ext_id['identifier']

                    # these "reactions" are what the drug is metabolized into
                    #     print(row['db_id, drug['reactions'])

        db_main.append(main_row)

        # SNP-FX contains data on the drug, the interacting protein(kegg),
        # the ‘causal’ SNPs or genetic variants for that gene/protein,
        # the therapeutic response or effects caused by the SNP-drug interaction
        # (improved or diminished response, changed dosing requirements, etc.)
        # and the associated references describing these effects in more detail.
        # SNP-ADR follows a similar format to SNP-FX but the clinical responses
        # are restricted only to adverse drug reactions (ADR).
        # SNP-FX contains literature-derived data on the therapeutic effects or therapeutic responses
        # for more than 60 drug-polymorphism combinations,
        # while SNP-ADR contains data on adverse reactions compiled from more than 50 drug-polymorphsim pairings.
        # All of the execute_dir in these tables is hyperlinked to drug entries from DrugBank, protein execute_dir from UniProt,
        # SNP execute_dir from dbSNP and bibliographic execute_dir from PubMed.
        # NOTE: these dataframes are combined, with a db_snp_adr column indicating a known ADR
        if drug['snp-effects']:
            for snp_eff in utils.force_list(drug['snp-effects']['effect']):
                snp_row = collections.OrderedDict()

                snp_row['db_id'] = db_id
                snp_row['db_snp_uniprot'] = snp_eff['uniprot-id']
                snp_row['db_snp_rsid'] = snp_eff['rs-id']
                snp_row['db_snp_allele'] = snp_eff['allele']
                snp_row['db_snp_allele_info'] = snp_eff['defining-change']
                snp_row['db_snp_effect'] = snp_eff['description']
                snp_row['db_snp_adr'] = False

                db_snp.append(snp_row)

        if drug['snp-adverse-drug-reactions']:
            for snp_eff in utils.force_list(drug['snp-adverse-drug-reactions']['reaction']):
                snp_row = collections.OrderedDict()

                snp_row['db_id'] = db_id
                snp_row['db_snp_uniprot'] = snp_eff['uniprot-id']
                snp_row['db_snp_rsid'] = snp_eff['rs-id']
                snp_row['db_snp_allele'] = snp_eff['allele']
                snp_row['db_snp_allele_info'] = snp_eff['adverse-reaction']
                snp_row['db_snp_effect'] = snp_eff['description']
                snp_row['db_snp_adr'] = True

                db_snp.append(snp_row)

        for bind_type in ['targets', 'enzymes', 'carriers', 'transporters']:
            if drug[bind_type]:
                for bind_to in utils.force_list(drug[bind_type][bind_type.strip('kegg')]):
                    if 'polypeptide' in bind_to.keys():
                        for polypeptide in utils.force_list(bind_to['polypeptide']):
                            bind_to_row = collections.OrderedDict()
                            bind_to_row['db_id'] = db_id
                            bind_to_row['db_uniprot'] = polypeptide['@id']
                            bind_to_row['db_type'] = bind_type.strip('kegg')
                            bind_to_row['db_known_action'] = bind_to['known-action']
                            bind_to_row['db_location'] = polypeptide['cellular-location']
                            if bind_to['actions']:
                                bind_to_row['db_action'] = utils.force_string(bind_to['actions']['action'])

                            db_bind_to.append(bind_to_row)

    pd.DataFrame(db_main).to_csv(db_main_path)
    pd.DataFrame(db_bind_to).to_csv(db_bind_path)
    pd.DataFrame(db_snp).to_csv(db_snp_path)

    print('Saved parsed dataframes: {}, {}, {}'.format(db_main_path, db_bind_path, db_snp_path))
    return db_main_path, db_bind_path, db_snp_path

if __name__ == '__main__':
    import argparse

    p = argparse.ArgumentParser(description='DrugBank tools')
    p.add_argument('user', help='DrugBank user name')
    p.add_argument('passwd', help='DrugBank password')
    p.add_argument('--outdir', '-o', help='Directory to write to')
    args = p.parse_args()

    print('Downloading DrugBank XML...')
    db_xml_path = download_drugbank_db(args.user, args.passwd, args.outdir)
    print('Parsing DrugBank database into dataframes...')
    parse_drugbank_xml(db_xml_path, outdir=args.outdir)


from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.PDBIO import Select
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.mmtf import MMTFParser
from Bio.PDB.PDBExceptions import PDBConstructionWarning
import os.path as op
import logging
import warnings
import ssbio.utils
from ssbio.biopython.bp_mmcifparser import MMCIFParserFix

log = logging.getLogger(__name__)

cifp = MMCIFParserFix(QUIET=True)
pdbp = PDBParser(PERMISSIVE=True, QUIET=True)
mmtfp = MMTFParser()


def as_protein(structure, filter_residues=True):
    """ Exposes methods in the Bio.Struct.Protein module.
        Parameters:
            - filter_residues boolean; removes non-aa residues through Bio.PDB.Polypeptide is_aa function
              [Default: True]
        Returns a new structure object.
    """

    from ssbio.biopython.Bio.Struct.Protein import Protein
    return Protein.from_structure(structure, filter_residues)


class StructureIO(PDBIO):
    """Extended class to load any structure file into a Biopython Structure object

    Loads the first model when there are multiple available.
    Also adds some logging methods.
    """

    def __init__(self, structure_file):
        PDBIO.__init__(self)

        dirname, filename_without_extension, file_type = ssbio.utils.split_folder_and_path(structure_file)
        self.structure_file = structure_file

        # Unzip the file if it is zipped
        if file_type == '.gz':
            unzipped = ssbio.utils.gunzip_file(structure_file, outdir=dirname)
            dirname, filename_without_extension, file_type = ssbio.utils.split_folder_and_path(unzipped)

        if file_type in ['.pdb', '.ent', '.mmcif', '.cif', '.mmtf']:
            # Load the structure
            if file_type.lower() == '.pdb' or file_type.lower() == '.ent':
                structure = pdbp.get_structure(id='ssbio_pdb', file=structure_file)
            if file_type.lower() == '.mmcif' or file_type.lower() == '.cif':
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', PDBConstructionWarning)
                    structure = cifp.get_structure(structure_id='ssbio_cif', filename=structure_file)
            if file_type.lower() == '.mmtf':
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', PDBConstructionWarning)
                    structure = mmtfp.get_structure(file_path=structure_file)
            log.debug('{}: parsed 3D coordinates of structure'.format(op.basename(structure_file)))
        else:
            raise ValueError('{}: unsupported file type'.format(file_type))

        # If there are multiple models (NMR), use the first model as the representative structure
        if len(structure) > 1:
            self.first_model = structure[0]
            structure = structure[0]
            self.set_structure(structure)
            log.debug('{}: using first model'.format(structure_file))
        elif len(structure) == 0:
            log.error('{}: no models in structure!'.format(structure_file))
        else:
            self.set_structure(structure)
            self.first_model = structure[0]

    def write_pdb(self, custom_name='', out_suffix='', out_dir=None, custom_selection=None, force_rerun=False):
        """Write a new PDB file for the Structure's FIRST MODEL.

        Set custom_selection to a PDB.Select class for custom SMCRA selections.

        Args:
            custom_name: Filename of the new file (without extension)
            out_suffix: Optional string to append to new PDB file
            out_dir: Optional directory to output the file
            custom_selection: Optional custom selection class
            force_rerun: If existing file should be overwritten

        Returns:
            out_file: filepath of new PDB file

        """
        if not custom_selection:
            custom_selection = Select()

        # If no output directory, custom name, or suffix is specified, add a suffix "_new"
        if not out_dir or not custom_name:
            if not out_suffix:
                out_suffix = '_new'

        # Prepare the output file path
        outfile = ssbio.utils.outfile_maker(inname=self.structure_file,
                                            outname=custom_name,
                                            append_to_name=out_suffix,
                                            outdir=out_dir,
                                            outext='.pdb')
        try:
            if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
                self.save(outfile, custom_selection)
        except TypeError:
            # If trying to save something that can't be saved as a PDB (example: 5iqr.cif), log an error and return None
            # The error thrown by PDBIO.py is "TypeError: %c requires int or char"
            log.error('{}: unable to save structure in PDB file format'.format(self.structure_file))
            return None

        return outfile


import unittest
import six

import ssbio.databases.uniprot
from ssbio.databases.uniprot import UniProtProp

class TestUniProt(unittest.TestCase):
    """Unit tests for UniProt
    """

    def test_uniprotprop(self):
        up = UniProtProp(uniprot_acc='P0ABP8', sequence_path='test_files/sequences/P0ABP8.fasta',
                         metadata_path='test_files/sequences/P0ABP8.txt')

        up_keys = ['bigg','description','ec_number','entry_version','gene_name','id','kegg',
                   'metadata_file','metadata_path','pdbs','pfam','refseq','reviewed','sequence_file','seq_len',
                   'sequence_path','seq_version','uniprot']

        for k in up_keys:
            # print(k)
            self.assertTrue(hasattr(up, k))

        self.assertEqual(up.bigg, None)
        self.assertEqual(up.description, ['Inosine phosphorylase', 'PNP', 'Purine nucleoside phosphorylase DeoD-type'])
        self.assertEqual(up.ec_number, ['2.4.2.1'])
        self.assertEqual(up.entry_version, '2016-11-02')
        self.assertEqual(up.gene_name, 'deoD')
        self.assertEqual(up.id, 'P0ABP8')
        six.assertCountEqual(self, up.kegg, ['ecj:JW4347', 'eco:b4384'])
        self.assertEqual(up.metadata_file, 'P0ABP8.txt')
        six.assertCountEqual(self, up.pdbs,
                        ['1OUM', '4TTA', '4TS3', '1OTY', '1K9S', '3OOE', '1OVG', '3ONV', '3UT6', '1OV6', '4TTJ', '1OTX',
                         '1OU4', '3OOH', '4TS9', '1A69', '3OPV', '1ECP', '4TTI'])
        self.assertEqual(up.pfam, ['PF01048'])
        six.assertCountEqual(self, up.refseq, ['NP_418801.1', 'NC_000913.3', 'WP_000224877.1', 'NZ_LN832404.1'])
        self.assertEqual(up.reviewed, True)
        self.assertEqual(up.sequence_file, 'P0ABP8.fasta')
        self.assertEqual(up.seq_len, 239)
        self.assertEqual(up.seq_version, '2007-01-23')
        self.assertEqual(up.uniprot, 'P0ABP8')

    def test_uniprot_valid_id(self):
        valid_ids = ['P20020',
                     'V9HWB9',
                     'A0A024QZK7',
                     'Q9Y633',
                     'O75414',
                     'B7Z3W8',
                     'V9HWJ2',
                     'F2Z2Y4',
                     'A0A024RDQ9',
                     'P48426']
        invalid_ids = ['1AAG', '1AN3', '1A8X', '8CEL', 'ASDASDASD##']

        for u in valid_ids:
            self.assertTrue(ssbio.databases.uniprot.is_valid_uniprot_id(u))

        for u in invalid_ids:
            self.assertFalse(ssbio.databases.uniprot.is_valid_uniprot_id(u))

    def test_uniprot_reviewed_checker(self):
        status = {'I1Z9G4': False,
                  'O14735': True}

        for f,v in status.items():
            self.assertEqual(ssbio.databases.uniprot.uniprot_reviewed_checker(f), v)

    def test_uniprot_reviewed_checker_batch(self):
        status = {'I1Z9G4': False,
                  'I3L3I9': False,
                  'J3KN10': False,
                  'J3KQV8': False,
                  'J3KRC4': False,
                  'O00408': True,
                  'O00757': True,
                  'O00764': True,
                  'O14494': True,
                  'O14495': True,
                  'O14735': True}

        test_status = ssbio.databases.uniprot.uniprot_reviewed_checker_batch(list(status.keys()))
        self.assertEqual(test_status, status)

    # def test_uniprot_metadata(self):
    #     test_out = {
    #         'P36959': {'u_description': ['Guanosine monophosphate reductase 1 {ECO:0000255|HAMAP-Rule:MF_03195}',
    #                                      "Guanosine 5'-monophosphate oxidoreductase 1 {ECO:0000255|HAMAP-Rule:MF_03195}",
    #                                      'GMPR 1 {ECO:0000255|HAMAP-Rule:MF_03195}',
    #                                      'GMP reductase 1 {ECO:0000255|HAMAP-Rule:MF_03195}'],
    #                    'u_ec_number': ['1.7.1.7'],
    #                    'u_entry_version': '2016-04-13',
    #                    'u_gene_name': 'GMPR',
    #                    'u_go': ['GO:0005829; C:cytosol; TAS:Reactome.',
    #                             'GO:1902560; C:GMP reductase complex; IEA:UniProtKB-EC.',
    #                             'GO:0003920; F:GMP reductase activity; TAS:ProtInc.',
    #                             'GO:0046872; F:metal ion binding; IEA:UniProtKB-KW.',
    #                             'GO:0055086; P:nucleobase-containing small molecule metabolic process; TAS:Reactome.',
    #                             'GO:0006144; P:purine nucleobase metabolic process; TAS:Reactome.',
    #                             'GO:0006163; P:purine nucleotide metabolic process; IEA:UniProtKB-HAMAP.',
    #                             'GO:0043101; P:purine-containing compound salvage; TAS:Reactome.',
    #                             'GO:0009409; P:response to cold; TAS:ProtInc.',
    #                             'GO:0044281; P:small molecule metabolic process; TAS:Reactome.'],
    #                    'u_kegg_id': ['hsa:2766'],
    #                    'u_pfam': ['PF00478'],
    #                    'u_refseq': ['NP_006868.3', 'NM_006877.3'],
    #                    'u_reviewed': True,
    #                    'u_seq': 'MPRIDADLKLDFKDVLLRPKRSSLKSRAEVDLERTFTFRNSKQTYSGIPIIVANMDTVGTFEMAAVMSQHSMFTAIHKHYSLDDWKLFATNHPECLQNVAVSSGSGQNDLEKMTSILEAVPQVKFICLDVANGYSEHFVEFVKLVRAKFPEHTIMAGNVVTGEMVEELILSGADIIKVGVGPGSVCTTRTKTGVGYPQLSAVIECADSAHGLKGHIISDGGCTCPGDVAKAFGAGADFVMLGGMFSGHTECAGEVFERNGRKLKLFYGMSSDTAMNKHAGGVAEYRASEGKTVEVPYKGDVENTILDILGGLRSTCTYVGAAKLKELSRRATFIRVTQQHNTVFS',
    #                    'u_seq_len': 345,
    #                    'u_seq_version': '1994-06-01',
    #                    'u_uniprot_acc': 'P36959'}}
    #
    #     self.assertEqual(ssbio.databases.uniprot.uniprot_metadata('P36959'), test_out)


import os.path as op
import unittest

from ssbio.protein.structure.structprop import StructProp

from ssbio.core.protein import Protein
from ssbio.databases.kegg import KEGGProp
from ssbio.databases.pdb import PDBProp
from ssbio.databases.uniprot import UniProtProp
from ssbio.protein.sequence.seqprop import SeqProp
from ssbio.protein.structure.homology.itasser.itasserprop import ITASSERProp


class TestProtein(unittest.TestCase):
    """Unit tests for Protein"""

    @classmethod
    def setUpClass(self):
        self.prot = Protein(ident='b4384')

    def test_load_kegg(self):
        new_kegg = self.prot.load_kegg('eco:b4384',
                                       kegg_seq_file='test_files/sequences/eco-b4384.faa',
                                       kegg_metadata_file='test_files/sequences/eco-b4384.kegg')
        self.assertTrue(self.prot.sequences.has_id('eco:b4384'))
        self.assertTrue(isinstance(new_kegg, KEGGProp))

    def test_load_uniprot(self):
        new_uniprot = self.prot.load_uniprot('P0ABP8',
                                             uniprot_seq_file='test_files/sequences/P0ABP8.fasta',
                                             uniprot_metadata_file='test_files/sequences/P0ABP8.txt')
        self.assertTrue(self.prot.sequences.has_id('P0ABP8'))
        self.assertTrue(isinstance(new_uniprot, UniProtProp))

    def test_filter_sequences(self):
        self.prot.load_kegg('keggdummy')
        self.prot.load_uniprot('P0OOR0')

        only_kegg = self.prot.filter_sequences(KEGGProp)
        for k in only_kegg:
            self.assertTrue(isinstance(k, KEGGProp))

        only_uniprot = self.prot.filter_sequences(UniProtProp)
        for k in only_uniprot:
            self.assertTrue(isinstance(k, UniProtProp))

    def test_load_manual_sequence_file(self):
        new_manual = self.prot.load_manual_sequence_file('tester', 'test_files/sequences/P0ABP8.fasta')
        self.assertTrue(self.prot.sequences.has_id('tester'))
        self.assertTrue(isinstance(new_manual, SeqProp))

    def test_load_manual_sequence(self):
        new_manual = self.prot.load_manual_sequence(ident='tester2', seq='ALALALAL', outdir='test_files/out/')
        self.assertTrue(self.prot.sequences.has_id('tester2'))
        self.assertTrue(isinstance(new_manual, SeqProp))

    def test_set_representative_sequence(self):
        self.prot.set_representative_sequence()
        self.assertEqual(self.prot.representative_sequence.id, 'P0ABP8')
        self.assertTrue(self.prot.representative_sequence.num_pdbs > 0)

    def test_align_sequences_to_representative(self):
        # TODO: unittests for this method
        # this is only true when you dont add more sequences
        # self.assertEqual(len(self.prot.representative_sequence.alignments), len(self.prot.sequences))
        pass

    def test_load_pdb(self):
        new_uniprot = self.prot.load_uniprot('P0ABP8',
                                             uniprot_seq_file='test_files/sequences/P0ABP8.fasta',
                                             uniprot_metadata_file='test_files/sequences/P0ABP8.txt',
                                             set_as_representative=True)

        newpdb = self.prot.load_pdb('1ecp', pdb_file='test_files/structures/1ecp.pdb', file_type='pdb')
        self.assertTrue(self.prot.structures.has_id('1ecp'))
        self.assertTrue(isinstance(newpdb, PDBProp))

    def test_load_itasser_folder(self):
        self.assertRaises(OSError, self.prot.load_itasser_folder, 'sdfsdfsdf', itasser_folder='sdfsdfsdf/structures/P9WG73')

        newitasser = self.prot.load_itasser_folder('P9WG73', itasser_folder='test_files/structures/P9WG73')
        self.assertTrue(self.prot.structures.has_id('P9WG73'))
        self.assertTrue(isinstance(newitasser, ITASSERProp))
        self.assertEqual(newitasser.structure_file, 'model1.pdb')

        # newitasser.copy_results(copy_to_dir='test_files/structures/test_out/', rename_model_to='haha', force_rerun=True)

    def test_load_homology_model(self):
        newprot = self.prot.load_generic_structure('DEOD', structure_file='test_files/structures/DEOD_ECOLI_model1.pdb')
        self.assertTrue(self.prot.structures.has_id('DEOD'))
        self.assertTrue(isinstance(newprot, StructProp))

    def test_num_structures(self):
        self.assertNotEqual(self.prot.num_structures, 0)

    def test_num_structures_experimental(self):
        self.assertNotEqual(self.prot.num_structures_experimental, 0)

    def test_num_structures_homology(self):
        self.assertNotEqual(self.prot.num_structures_homology, 0)

    def test_set_representative_structure(self):
        prot_to_set = self.prot.structures.get_by_id('1ecp')
        self.prot.set_representative_structure(seq_outdir=op.join('test_files','out'),
                                               struct_outdir=op.join('test_files','out'),
                                               pdb_file_type='cif')
        self.assertEqual('1ecp-A', self.prot.representative_structure.id)


import os.path as op
import unittest

from ssbio.protein.structure.utils.cleanpdb import CleanPDB
from ssbio.protein.structure.utils.structureio import StructureIO


class TestCleanPDB(unittest.TestCase):
    """Unit tests for CleanPDB
    """

    def test_clean_pdb(self):
        files = [('1kf6.pdb', '1kf6_clean_tester.pdb'), ('PHOE_ECOLI_model1.pdb', 'PHOE_ECOLI_model1_clean_tester.pdb'),
                 ('E04142.pdb', 'E04142_clean_tester.pdb'), ('1cbn.pdb', '1cbn_clean_tester.pdb')]

        outdir = op.join('test_files', 'out')
        working_dir = 'test_structures'
        out_suffix = '_clean'
        custom_clean = CleanPDB()

        for infile,outfile in files:
            outfile_new = '{}{}.pdb'.format(op.splitext(infile)[0], out_suffix)
            infile_path = op.join(working_dir, infile)

            my_pdb = StructureIO(infile_path)
            default_cleaned_pdb = my_pdb.write_pdb(custom_selection=custom_clean, out_suffix=out_suffix,
                                                   out_dir=outdir, force_rerun=True)
            default_cleaned_pdb_basename = op.basename(default_cleaned_pdb)

            # test if the filename is correct
            self.assertEqual(default_cleaned_pdb_basename, outfile_new)

            # test if the file contents are equal
            self.assertEqual(open(default_cleaned_pdb,'r').read(),
                             open(op.join(working_dir, outfile), 'r').read())

            # test that the file does not equal the original file
            self.assertNotEqual(open(default_cleaned_pdb,'r').read(),
                                open(infile_path, 'r').read())

    def test_clean_pdb_and_get_chain(self):
        files = [('1kf6.pdb', '1kf6_clean_chainA_tester.pdb')]
        outdir = op.join('test_files', 'out')
        working_dir = 'test_structures'
        out_suffix = '_clean_chainA'
        custom_clean = CleanPDB(keep_chains='A')

        for infile, outfile in files:
            outfile_new = '{}{}.pdb'.format(op.splitext(infile)[0], out_suffix)
            infile_path = op.join(working_dir, infile)

            my_pdb = StructureIO(infile_path)
            default_cleaned_pdb = my_pdb.write_pdb(custom_selection=custom_clean, out_suffix=out_suffix,
                                                   out_dir=outdir, force_rerun=True)
            default_cleaned_pdb_basename = op.basename(default_cleaned_pdb)

            # test if the filename is correct
            # print(default_cleaned_pdb_basename, outfile_new)
            self.assertEqual(default_cleaned_pdb_basename, outfile_new)

            # test if the file contents are equal
            self.assertEqual(open(default_cleaned_pdb, 'r').read(),
                             open(op.join(working_dir, outfile), 'r').read())

            # test that the file does not equal the original file
            self.assertNotEqual(open(default_cleaned_pdb, 'r').read(),
                                open(infile_path, 'r').read())


import logging
from collections import defaultdict
from copy import deepcopy

from Bio.Alphabet import IUPAC
from Bio.PDB import Polypeptide
from Bio.PDB.HSExposure import ExposureCN, HSExposureCA, HSExposureCB
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

import ssbio.protein.sequence.utils
from ssbio.protein.structure.utils.structureio import StructureIO

log = logging.getLogger(__name__)


def search_ss_bonds(model, threshold=3.0):
    """ Searches S-S bonds based on distances
        between atoms in the structure (first model only).
        Average distance is 2.05A. Threshold is 3A default.
        Returns iterator with tuples of residues.

        # ADAPTED FROM JOAO RODRIGUES BIOPYTHON GSOC PROJECT
    """

    # Taken from http://docs.python.org/library/itertools.html
    # Python 2.4 does not include itertools.combinations

    def combinations(iterable, r):
        # combinations('ABCD', 2) --> AB AC AD BC BD CD
        # combinations(range(4), 3) --> 012 013 023 123
        pool = tuple(iterable)
        n = len(pool)
        if r > n:
            return
        indices = list(range(r))
        yield tuple(pool[i] for i in indices)
        while True:
            for i in reversed(range(r)):
                if indices[i] != i + n - r:
                    break
            else:
                return
            indices[i] += 1
            for j in range(i + 1, r):
                indices[j] = indices[j - 1] + 1
            yield tuple(pool[i] for i in indices)

    cysteines = [r for r in model.get_residues() if r.get_resname() == 'CYS']

    pairs = combinations(cysteines, 2)  # Iterator with pairs

    bridges = []
    for cys_pair in pairs:
        if cys_pair[0]['SG'] - cys_pair[1]['SG'] < threshold:
            bridges.append(cys_pair)

    infodict = {}
    if bridges:
        infodict = defaultdict(list)

        for disulfide_bridge in bridges:
            residue1 = disulfide_bridge[0]
            residue2 = disulfide_bridge[1]
            chain = residue1.get_parent().id
            infodict[chain].append((residue1.get_full_id()[3], residue2.get_full_id()[3]))

    return infodict


def get_structure_seqrecords(model):
    """Get a dictionary of a PDB file's sequences.

    Special cases include:
        - Insertion codes. In the case of residue numbers like "15A", "15B", both residues are written out. Example: 9LPR
        - HETATMs. Currently written as an "X", or unknown amino acid.

    Args:
        model: Biopython Model object of a Structure

    Returns:
        list: List of SeqRecords

    """

    structure_seq_records = []

    # Loop over each chain of the PDB
    for chain in model:
        tracker = 0
        chain_seq = ''
        chain_resnums = []

        # Loop over the residues
        for res in chain.get_residues():
            # NOTE: you can get the residue number too
            res_id = res.id
            res_num = res_id[1]
            res_icode = res_id[2]

            # Double check if the residue name is a standard residue
            # If it is not a standard residue (ie. selenomethionine),
            # it will be filled in with an X on the next iteration)
            if Polypeptide.is_aa(res, standard=True):
                end_tracker = res_num
                res_aa_one = Polypeptide.three_to_one(res.get_resname())

                # Tracker to fill in X's
                if end_tracker != (tracker + 1):
                    if res_icode != ' ':
                        chain_seq += res_aa_one
                        chain_resnums.append(res_id)
                        tracker = end_tracker + 1
                        continue
                    else:
                        multiplier = (end_tracker - tracker - 1)
                        chain_seq += 'X' * multiplier
                        # Residue numbers for unresolved or nonstandard residues are Infinite
                        chain_resnums.extend([(' ', float("Inf"), ' ')] * multiplier)

                chain_seq += res_aa_one
                chain_resnums.append(res_id)
                tracker = end_tracker

            else:
                continue

        chain_seq_record = SeqRecord(Seq(chain_seq, IUPAC.protein), id=chain.get_id())
        chain_seq_record.letter_annotations['structure_resnums'] = chain_resnums
        structure_seq_records.append(chain_seq_record)

    return structure_seq_records


def get_structure_seqs(pdb_file, file_type):
    """Get a dictionary of a PDB file's sequences.

    Special cases include:
        - Insertion codes. In the case of residue numbers like "15A", "15B", both residues are written out. Example: 9LPR
        - HETATMs. Currently written as an "X", or unknown amino acid.

    Args:
        pdb_file: Path to PDB file

    Returns:
        dict: Dictionary of:
        {chain_id: sequence}

    """

    # TODO: Please check out capitalization of chain IDs in mmcif files. example: 5afi - chain "l" is present but
    # it seems like biopython capitalizes it to chain L

    # Get the first model
    my_structure = StructureIO(pdb_file)
    model = my_structure.first_model

    structure_seqs = {}

    # Loop over each chain of the PDB
    for chain in model:
        chain_seq = ''
        tracker = 0

        # Loop over the residues
        for res in chain.get_residues():
            # NOTE: you can get the residue number too
            # res_num = res.id[1]

            # Double check if the residue name is a standard residue
            # If it is not a standard residue (ie. selenomethionine),
            # it will be filled in with an X on the next iteration)
            if Polypeptide.is_aa(res, standard=True):
                full_id = res.get_full_id()
                end_tracker = full_id[3][1]
                i_code = full_id[3][2]
                aa = Polypeptide.three_to_one(res.get_resname())

                # Tracker to fill in X's
                if end_tracker != (tracker + 1):
                    if i_code != ' ':
                        chain_seq += aa
                        tracker = end_tracker + 1
                        continue
                    else:
                        chain_seq += 'X' * (end_tracker - tracker - 1)

                chain_seq += aa
                tracker = end_tracker

            else:
                continue

        structure_seqs[chain.get_id()] = chain_seq

    return structure_seqs


def match_structure_sequence(orig_seq, new_seq, match='X', fill_with='X'):
    """Correct a sequence to match inserted X's in a structure sequence

    This is useful for mapping a sequence obtained from structural tools like MSMS or DSSP
        to the sequence obtained by the get_structure_seqs method.

    Examples:
        >>> structure_seq = 'XXXABCDEF'
        >>> prop_list = [4, 5, 6, 7, 8, 9]
        >>> match_structure_sequence(structure_seq, prop_list)
        ['X', 'X', 'X', 4, 5, 6, 7, 8, 9]

        >>> match_structure_sequence(structure_seq, prop_list, fill_with=float('Inf'))
        [inf, inf, inf, 4, 5, 6, 7, 8, 9]

        >>> structure_seq = '---ABCDEF---'
        >>> prop_list = ('H','H','H','C','C','C')
        >>> match_structure_sequence(structure_seq, prop_list, match='-', fill_with='-')
        ('-', '-', '-', 'H', 'H', 'H', 'C', 'C', 'C', '-', '-', '-')

        >>> structure_seq = 'ABCDEF---'
        >>> prop_list = 'HHHCCC'
        >>> match_structure_sequence(structure_seq, prop_list, match='-', fill_with='-')
        'HHHCCC---'

        >>> structure_seq = 'AXBXCXDXEXF'
        >>> prop_list = ['H', 'H', 'H', 'C', 'C', 'C']
        >>> match_structure_sequence(structure_seq, prop_list, match='X', fill_with='X')
        ['H', 'X', 'H', 'X', 'H', 'X', 'C', 'X', 'C', 'X', 'C']

    Args:
        orig_seq (str, Seq, SeqRecord): Sequence to match to
        new_seq (str, tuple, list): Sequence to fill in
        match (str): What to match
        fill_with: What to fill in when matches are found

    Returns:
        str, tuple, list: new_seq which will match the length of orig_seq

    """
    if len(orig_seq) == len(new_seq):
        log.debug('Lengths already equal, nothing to fill in')
        return new_seq

    if len(orig_seq) < len(new_seq):
        raise ValueError('Original sequence has a length less than the sequence provided to match to')

    if not isinstance(new_seq, str) and not isinstance(new_seq, tuple) and not isinstance(new_seq, list):
        raise ValueError('Invalid sequence provided, must be string, tuple, or list')

    orig_seq = ssbio.protein.sequence.utils.cast_to_str(orig_seq)
    new_thing = deepcopy(new_seq)
    if isinstance(new_seq, tuple):
        new_thing = list(new_thing)

    for i, s in enumerate(orig_seq):
        if s == match:
            if isinstance(new_thing, str):
                new_thing = new_thing[:i] + fill_with + new_thing[i:]
            if isinstance(new_thing, list):
                new_thing.insert(i, fill_with)

    if isinstance(new_seq, tuple):
        new_thing = tuple(new_thing)

    return new_thing


def site_centroid(residues, model):
    """Get the XYZ coordinate of the center of a list of residues.

    Args:
        residues: List of residue numbers
        pdb_file: Path to PDB file

    Returns:
        tuple: (X, Y, Z) coordinate of centroid

    """
    pass


def distance_to_site(residue_of_interest, residues, model):
    """Calculate the distance between an amino acid and a group of amino acids.

    Args:
        residue_of_interest: Residue number you are interested in (ie. a mutation)
        residues: List of residue numbers

    Returns:
        float: Distance (in Angstroms) to the group of residues

    """
    centroid = site_centroid(residues, residue_of_interest)
    pass


# TODO: half sphere exposure
def hse_output(pdb_file, file_type):
    """
    The solvent exposure of an amino acid residue is important for analyzing,
    understanding and predicting aspects of protein structure and function [73].
    A residue's solvent exposure can be classified as four categories: exposed, partly exposed,
    buried and deeply buried residues. Hamelryck  et al. [73] established a new 2D measure that provides a
    different view of solvent exposure, i.e. half-sphere exposure (HSE). By conceptually dividing the sphere
    of a residue into two halves- HSE-up and HSE-down, HSE provides a more detailed description of an amino
    acid residue's spatial neighborhood. HSE is calculated by the hsexpo module implemented in the BioPython
    package [74] from a PDB file.

    http://onlinelibrary.wiley.com/doi/10.1002/prot.20379/abstract

    Args:
        pdb_file:

    Returns:

    """
    # Get the first model
    my_structure = StructureIO(pdb_file)
    model = my_structure.first_model

    # Calculate HSEalpha
    exp_ca = HSExposureCA(model)
    # Calculate HSEbeta
    exp_cb = HSExposureCB(model)
    # Calculate classical coordination number
    exp_fs = ExposureCN(model)

    return
# def magni(a, b, c):
#     """Calculate the magnitude of distance vector
#     """
#     return pow((pow(a, 2) + pow(b, 2) + pow(c, 2)), 1.0 / 2.0)


# @cachetools.func.ttl_cache(maxsize=256)
# def calculate_res_distance(res_1, res_2, pdb_file):
#     """Calculate distance of one residue number to another in a PDB file
#
#     Args:
#         res_1: Residue number 1
#         res_2: Residue number 2
#         pdb_file: Path to PDB file
#
#     Returns:
#
#     """
#
#     my_structure = StructureIO(pdb_file)
#     model = my_structure.first_model
#
#     res_list = PDB.Selection.unfold_entities(model, 'R')
#
#     ires_list = []
#     res_chk_1 = ''
#     res_chk_2 = ''
#     for j in res_list:
#         if j.id[1] in [res_1, res_2] and j.resname != 'HOH':
#             ires_list.append(j)
#             if res_chk_1 == '' and res_chk_2 == '':
#                 res_chk_1 = j.id[1]
#             else:
#                 res_chk_2 = j.id[1]
#
#     paired = ssbio.utils.combinations(ires_list, 2)
#     try:
#         for k in paired:
#             chainA = PDB.Selection.unfold_entities(k[0], 'C')[0]
#             chainB = PDB.Selection.unfold_entities(k[1], 'C')[0]
#             vec = list(
#                 np.array([x.get_coord() for x in k[0]]).mean(axis=0) - np.array([x.get_coord() for x in k[1]]).mean(
#                     axis=0))
#             distance = magni(vec[0], vec[1], vec[2])
#
#         return distance
#     except UnboundLocalError:
#         log.error("Unknown interaction")
#         return None


import ftplib
import os.path as op
import logging
import os
log = logging.getLogger(__name__)


def download_coding_sequences(patric_id, seqtype, outdir='', outfile='', force_rerun=False):
    """Download the entire set of DNA or protein sequences from protein-encoding genes in a genome from NCBI.

    Saves a FASTA file in the optional directory specified.

    Args:
        genome_accession_or_id (str): PATRIC ID
        seqtype (str): "dna" or "protein" - if you want the coding sequences in DNA or amino acid formats.
        outdir (str): optional output directory (default is the current directory)
        outfile (str): optional custom name for file
        force_rerun (bool): if you want to redownload existing files

    Returns:
        Path to downloaded FASTA file.

    """
    if seqtype == 'dna':
        extension = 'ffn'
    elif seqtype == 'protein':
        extension = 'faa'
    else:
        raise ValueError('seqtype must be "dna" or "protein"')

    # TODO: use utils functions here
    # path and filename parsing
    if outfile:
        outfile = op.join(outdir, '{}.{}'.format(outfile, extension))
    else:
        # if no outfile is specified, default is "$GI.PATRIC.faa"
        outfile = op.join(outdir, '{}.PATRIC.{}'.format(patric_id, extension))

    if not force_rerun:
        # return the path to the file if it was already downloaded
        if op.exists(outfile) and os.stat(outfile).st_size != 0:
            log.debug('FASTA file already exists at {}'.format(outfile))
            return outfile

    try:
        ftp = ftplib.FTP('ftp.patricbrc.org')
        ftp.login()
        ftp.cwd("/patric2/patric3/genomes/{0}/".format(patric_id))
        with open(outfile, "wb") as gFile:
            ftp.retrbinary('RETR {0}.PATRIC.{1}'.format(patric_id, extension), gFile.write)
        ftp.quit()
    # TODO: check exceptions
    except:
        return None

    return outfile


import re
import os.path as op
import cobra.io.json
from cobra.io import load_matlab_model
from cobra.io import read_sbml_model
from cobra.io import load_json_model
from cobra.core import Model
from cobra.core import Gene
from cobra.core import DictList
from ssbio.core.genepro import GenePro


class ModelPro(Model):
    def __new__(cls, *args, **kwargs):
        """Casting Gene objects into GenePro objects
            This replaces any new instance of Genes with GenePros. Even when you load a model later
            using COBRApy methods. Use with caution!
        See http://stackoverflow.com/questions/3464061/cast-base-class-to-derived-class-python-or-more-pythonic-way-of-extending-class

        Returns:
            GenePro: a Gene object with a .protein attribute
        """
        if cls == Gene:
            return object.__new__(GenePro)
        return object.__new__(cls)

    Gene.__new__ = staticmethod(__new__)

    def __json_encode__(self):
        """convert the model to a dict"""
        obj = dict(
                reactions=[cobra.io.json.reaction_to_dict(reaction) for reaction in self.reactions],
                metabolites=[
                    cobra.io.json.metabolite_to_dict(metabolite) for metabolite in self.metabolites
                    ],
                genes = self.genes,
                id=self.id,
        )
        cobra.io.json._update_optional(self, obj, cobra.io.json._OPTIONAL_MODEL_ATTRIBUTES)
        # add in the JSON version
        obj["version"] = 1
        return obj

    def __json_decode__(self, **attrs):
        """build a model from a dict"""
        Model.__init__(self)
        if 'reactions' not in attrs:
            raise Exception('JSON object has no reactions attribute. Cannot load.')
        self.add_metabolites(
                [cobra.io.json.metabolite_from_dict(metabolite) for metabolite in attrs['metabolites']]
        )
        self.genes = DictList(attrs['genes'])
        self.add_reactions(
                [cobra.io.json.reaction_from_dict(reaction, self) for reaction in attrs['reactions']]
        )
        for k, v in attrs.items():
            if k in {'id', 'name', 'notes', 'compartments', 'annotation'}:
                setattr(self, k, v)


def model_loader(gem_file_path, gem_file_type, pdb_file_type='cif'):
    """Consolidated function to load a GEM using COBRApy. Specify the file type being loaded.

    Args:
        gem_file_path (str): Path to model file
        gem_file_type (str): GEM model type - 'sbml' (or 'xml'), 'mat', or 'json' format

    Returns:
        COBRApy Model object.

    """

    extension = op.splitext(gem_file_path)[1]

    if gem_file_type.replace('.', '').lower() == 'xml' or gem_file_type.replace('.', '').lower() == 'sbml':
        model = read_sbml_model(gem_file_path)
    elif extension.replace('.', '').lower() == 'mat':
        model = load_matlab_model(gem_file_path)
    elif extension.replace('.', '').lower() == 'json':
        model = load_json_model(gem_file_path)
    else:
        raise ValueError('File type must be in SBML, MATLAB, or JSON formats.')

    modelpro = ModelPro(model)
    for g in modelpro.genes:
        g.protein.pdb_file_type = pdb_file_type

    return modelpro


def is_spontaneous(gene):
    """Input a COBRApy Gene object and check if the ID matches a spontaneous ID regex.

    :param gene:
    :return:
    """
    spont = re.compile("[Ss](_|)0001")
    if spont.match(gene.id):
        return True
    else:
        return False


def filter_out_spontaneous_genes(genes):
    """Return the DictList of genes that are not spontaneous in a model.

    Args:
        model: COBRApy Model object

    Returns:
        DictList: genes excluding ones that are spontaneous

    """
    new_genes = DictList()
    for gene in genes:
        if not is_spontaneous(gene):
            new_genes.append(gene)

    return new_genes

def true_num_genes(model):
    """Return the number of genes in a model ignoring spontaneously labeled genes

    :param model: COBRApy Model object
    :return int: Number of genes excluding spontaneous genes
    """
    true_num = 0
    for gene in model.genes:
        if not is_spontaneous(gene):
            true_num += 1
    return true_num


def true_num_reactions(model):
    true_num = 0
    for rxn in model.reactions:
        if len(rxn.genes) == 0:
            continue
        if len(rxn.genes) == 1 and is_spontaneous(list(rxn.genes)[0]):
            continue
        else:
            true_num += 1
    return true_num


def adj_num_reactions(model, missing_genes):
    adj_num = 0
    for rxn in model.reactions:
        if len(rxn.genes) == 0:
            continue
        if len(rxn.genes) == 1 and (is_spontaneous(list(rxn.genes)[0]) or list(rxn.genes)[0] in missing_genes):
            continue
        else:
            adj_num += 1
    return adj_num


#!/usr/bin/env python
from __future__ import print_function
import argparse
import glob
import os
import os.path as op
import pandas as pd
from tqdm import tqdm
from ssbio import utils
date = utils.Date()
try:
    from urllib.request import urlopen
    from urllib.request import build_opener
    from urllib.request import HTTPCookieProcessor
    from urllib.parse import urlparse
    from urllib.parse import urlencode
except ImportError:
    from urlparse import urlparse
    from urllib2 import urlopen
    from urllib import urlencode
    from urllib2 import build_opener
    from urllib2 import HTTPCookieProcessor
# TODO: replace urllib usage with six library
try:
    from http.cookiejar import CookieJar
except ImportError:
    from cookielib import CookieJar

import glob
import time
# import cachetools
import requests


# author: Ke Chen

class AMYLPRED:
    def __init__(self, email, password):
        self.email = email
        self.password = password

    def get_aggregation_propensity(self, seq, cutoff_v=5, cutoff_n=5):
        """Predict aggregation propensity

        Args:
            seq:     amino acid sequence
            cutoff_v: the minimal number of methods that agree on
                      a residue being a aggregation-prone residue
            cutoff_n: minimal number of consecutive residues to be
                      considered as a 'stretch' of aggregation-prone region

        Returns:
            aggregation propensity

        """

        output = self.consensus_aggregation(seq)
        agg_index, agg_conf = self.write_propensity(len(seq), output, cutoff_v=cutoff_v, cutoff_n=cutoff_n)

        return agg_index

    # @cachetools.func.ttl_cache(maxsize=128)
    def consensus_aggregation(self, seq):
        url = "http://aias.biol.uoa.gr/AMYLPRED2/login.php"
        cj = CookieJar()
        opener = build_opener(HTTPCookieProcessor(cj))
        formdata = {"email": self.email, "password": self.password}
        data_encoded = urlencode(formdata)
        data_encoded = data_encoded.encode('ASCII')
        response = opener.open(url, data_encoded)

        # TODO: usually AMYLMUTS is most time consuming,
        #        and generate a slightly different result every submission
        #        consider remove this from list to save computational time?
        #        will need redo statistics
        Methods = ['AGGRESCAN', 'NETCSSP', 'PAFIG', 'APD', 'AMYLPATTERN',
                   'SECSTR', 'BSC', 'WALTZ', 'CONFENERGY', 'TANGO']#, 'AMYLMUTS']

        # TODO: can each method be cached?


        output = {}
        timeCounts = 0
        for met in Methods:

            # first check if there is an existing results file
            existing_results = glob.glob('*_{}.txt'.format(met))
            if existing_results:
                results_file = existing_results[0]
            else:
                values = {'seq_data': seq, 'method': met}
                data = urlencode(values)
                data = data.encode('ASCII')
                # url_input = "http://aias.biol.uoa.gr/AMYLPRED2/input.php"
                url_input = "http://aias.biol.uoa.gr/cgi-bin/AMYLPRED2/amylpred2.pl"
                response = opener.open(url_input, data)
                result = str(response.read())
                ind = str.find(result, 'Job ID')
                result2 = result[ind:ind + 50]
                ind1 = str.find(result2, ':')
                ind2 = str.find(result2, '<BR>')
                job_id = result2[ind1 + 2:ind2]

                # waiting for the calculation to complete
                url_result = 'http://aias.biol.uoa.gr/AMYLPRED2/tmp/' + job_id + '.txt'
                print(url_result)
                print("Waiting for %s results" % met, end='.')
                while True:
                    result = urlopen(url_result).read()
                    if not result:
                        time.sleep(1)
                        timeCounts += 1
                        print('.', end='')
                    else:
                        response = requests.get(url_result)
                        break
                # TODO: utilize saved file as cached resultr
                results_file = "{}_{}.txt".format(url_result.split('/')[-1].strip('.txt'), met)
                with open(results_file, "wb") as handle:
                    for data in response.iter_content():
                        handle.write(data)
            print("")
            method, hits = self.get_method_hits(results_file, met)
            # if method.lower() == met.lower():
            output[met] = hits
            # elif method == 'Beta-strand contiguity' and met == 'BSC':
            # output[met]=hits
            # elif method == 'Hexapeptide Conf. Energy' and met == 'CONFENERGY':
        print("Time Spent: %d seconds" % timeCounts)
        return output

    def get_method_hits(self, results_file, met):
        result = str(open(results_file).read())
        ind_s = str.find(result, 'HITS')
        ind_e = str.find(result, '**NOTE')
        tmp = result[ind_s + 10:ind_e].strip(" ")
        hits_resid = []
        method = None
        if ":" in tmp:
            method = tmp.split(":")[0]
            hits = tmp.split(":")[1]
            if "-" in hits:
                for ele in hits.split(","):
                    ele = ele.replace('\\r\\n\\r\\n', '')
                    res_s = ele.split("-")[0]
                    res_e = ele.split("-")[1]
                    for i in range(int(res_s), int(res_e) + 1):
                        hits_resid.append(i)
        if method:
            return method, hits_resid
        else:
            return met, hits_resid

    def get_aggregation_index(self, output, cutoff_v, cutoff_n):
        # all index for all aggregation potential > cutoff_v
        IDX = [i for i in range(0, len(output)) if output[i] >= cutoff_v]
        # hold number of continueous aa with aggregation potential > cutoff_v
        # for each continueous stretch
        Agg_index = []
        IDX_for_summation = []
        if IDX:
            # start index for each stretch
            start_IDX = [IDX[0]]
            i = 0
            counter = 0
            while i < len(IDX) - 1:
                if IDX[i] + 1 == IDX[i + 1]:
                    counter += 1
                else:
                    Agg_index.append(counter + 1)
                    start_IDX.append(IDX[i + 1])
                    counter = 0
                i += 1
                if i == len(IDX) - 1: Agg_index.append(counter + 1)
            IDX2 = [i for i in range(0, len(Agg_index)) if Agg_index[i] >= cutoff_n]
            for i in IDX2:
                IDX_for_summation.extend(range(start_IDX[i], Agg_index[i] + start_IDX[i]))
            sum = 0
            for i in IDX_for_summation:
                sum += output[i]
            sum_all = 0
        else:
            IDX2 = []
            sum = 0
        if len(IDX_for_summation) == 0:
            return len(IDX2), 0
        else:
            return len(IDX2), sum / float(len(IDX_for_summation))

    def write_propensity(self, N, output, cutoff_v=5, cutoff_n=5):
        Hits_res_count = {}
        for met in output.keys():
            for res in output[met]:
                if res in Hits_res_count.keys():
                    Hits_res_count[res] += 1
                else:
                    Hits_res_count[res] = 1

        AGG_Propensity = []
        output2 = []
        for i in range(1, int(N) + 1):
            if i in Hits_res_count.keys():
                output2.append(Hits_res_count[i])
            else:
                output2.append(0.0)
        agg_index, agg_conf = self.get_aggregation_index(output2, cutoff_v,
                                                         cutoff_n)  # these are parameters that you can explore, "5,5" means
        # 5 consecutive residues agreed among at least 5 methods
        # is considered contributing 1 to the aggregation propensity
        return agg_index, agg_conf

if __name__ == '__main__':
    # load inputs from command line

    p = argparse.ArgumentParser(description='Run AMYLPRED2 on a FASTA file or a folder of FASTA files.')
    p.add_argument('email', help='http://aias.biol.uoa.gr/AMYLPRED2/login.php Email')
    p.add_argument('password', help='Password')
    p.add_argument('infile', help='FASTA file or directory of FASTA files.')
    args = p.parse_args()

    curr_dir = os.getcwd()
    # initialize the class with your email and password for the site
    agg_predictions = agg.AMYLPRED(args.email, args.password)

    prop_dir = 'properties'
    if not op.exists(prop_dir):
        os.mkdir(prop_dir)

    agg_prop_dir = op.join(prop_dir,'aggregation_propensity')
    if not op.exists(agg_prop_dir):
        os.mkdir(agg_prop_dir)

    # TODO: improve arg parsing for files/dirs
    # TODO: this was all done in a rush - current dir and infile should be improved
    if len(args.infile) == 1 and op.isdir(args.infile[0]):
        os.chdir(args.infile[0])
        files = glob.glob('*')
    else:
        files = args.infile

    results = []

    for file in tqdm(files):
        if op.isdir(file):
            continue

        # load the sequence file, also the ID
        seq_records = fasta.load_fasta_file(file)
        seq_id = op.splitext(op.basename(file))[0]

        seq_folder = op.join(agg_prop_dir, seq_id)
        if not op.exists(seq_folder):
            os.mkdir(seq_folder)

        os.chdir(seq_folder)
        # TODO: seems useless to return seqrecords to just convert them to strings
        for seq_record in seq_records:
            agg_index = agg_predictions.get_aggregation_propensity(str(seq_record.seq))
            result = {'id':seq_id, 'agg_index':agg_index}
            results.append(result)
        os.chdir(curr_dir)

    agg_df = pd.DataFrame(results)
    agg_df.to_csv(op.join(prop_dir, '{}_aggprop_results.csv'.format(date.short_date)))
    print('Saved results in properties/aggregation_propensity and summarized in aggprop_results.csv')


import logging
import os.path as op
import shutil
from collections import OrderedDict

import pandas as pd
import requests
from Bio.PDB.PDBExceptions import PDBConstructionException
from Bio.Seq import Seq
from cobra.core import DictList
from six.moves.urllib.error import URLError
from slugify import Slugify

import ssbio.databases.pdb
import ssbio.protein.sequence.utils.alignment
import ssbio.protein.sequence.utils.fasta
import ssbio.protein.structure.properties.quality
import ssbio.utils
from ssbio.core.object import Object
from ssbio.databases.kegg import KEGGProp
from ssbio.databases.pdb import PDBProp
from ssbio.databases.uniprot import UniProtProp
from ssbio.protein.sequence.seqprop import SeqProp
from ssbio.protein.structure.homology.itasser.itasserprop import ITASSERProp
from ssbio.protein.structure.structprop import StructProp

custom_slugify = Slugify(safe_chars='-_.')
log = logging.getLogger(__name__)


class Protein(Object):
    """Basic definition of a protein"""

    __representative_sequence_attributes = ['gene', 'uniprot', 'kegg', 'pdbs',
                                            'sequence_path', 'metadata_path']
    __representative_structure_attributes = ['is_experimental', 'reference_seq_top_coverage', 'date', 'description',
                                             'resolution','taxonomy_name']

    def __init__(self, ident, description=None, root_dir=None, pdb_file_type='cif'):
        """Initialize a Protein object.

        A Protein contains sequences, structures, and a single representative sequence and structure.

        Args:
            ident (str): Unique identifier for this protein
            description (str): Optional description for this protein
            root_dir:
            pdb_file_type:
        """
        Object.__init__(self, id=ident, description=description)
        self.sequences = DictList()
        self.structures = DictList()
        self.representative_sequence = None
        self.representative_structure = None

        self.root_dir = root_dir
        self.pdb_file_type = pdb_file_type

    @property
    def root_dir(self):
        """str: Directory where Protein project folder is located"""
        return self._root_dir

    @root_dir.setter
    def root_dir(self, path):
        self._root_dir = path

        if path:
            if not op.exists(path):
                raise ValueError('{}: folder does not exist'.format(path))

            for d in [self.protein_dir, self.sequence_dir, self.structure_dir]:
                ssbio.utils.make_dir(d)

            for seq in self.sequences:
                seq.sequence_dir = self.sequence_dir
                seq.metadata_dir = self.sequence_dir

            for struct in self.structures:
                struct.root_dir = self.structure_dir

    @property
    def protein_dir(self):
        """str: Protein folder"""
        if self.root_dir:
            # Add a _protein suffix to the folder if it has the same name as its root folder
            folder_name = self.id
            if folder_name == op.basename(self.root_dir):
                folder_name = self.id + '_protein'
            return op.join(self.root_dir, folder_name)
        else:
            log.warning('Root directory not set')
            return None

    @property
    def sequence_dir(self):
        """str: Directory where sequence related files are stored"""
        if self.root_dir:
            return op.join(self.protein_dir, 'sequences')
        else:
            log.debug('Root directory not set')
            return None

    @property
    def structure_dir(self):
        """str: Directory where structure related files are stored"""
        if self.root_dir:
            return op.join(self.protein_dir, 'structures')
        else:
            log.debug('Root directory not set')
            return None

    @property
    def num_structures(self):
        """int: Return the total number of structures"""
        return len(self.structures)

    @property
    def num_structures_experimental(self):
        """int: Return the total number of experimental structures"""
        return len(self.get_experimental_structures())

    @property
    def num_structures_homology(self):
        """int: Return the total number of homology models"""
        return len(self.get_homology_models())

    def get_experimental_structures(self):
        """DictList: Return a DictList of all experimental structures in self.structures"""
        return DictList(x for x in self.structures if x.is_experimental)

    def get_homology_models(self):
        """DictList: Return a DictList of all homology models in self.structures"""
        return DictList(x for x in self.structures if not x.is_experimental)

    def filter_sequences(self, seq_type):
        """Return a DictList of only specified types in the sequences attribute.

        Args:
            seq_type (SeqProp): Object type

        Returns:
            DictList: A filtered DictList of specified object type only

        """
        return DictList(x for x in self.sequences if isinstance(x, seq_type))

    def load_kegg(self, kegg_id, kegg_organism_code=None, kegg_seq_file=None, kegg_metadata_file=None,
                  set_as_representative=False, download=False, outdir=None, force_rerun=False):
        """Load a KEGG ID, sequence, and metadata files into the sequences attribute.

        Args:
            kegg_id (str): KEGG ID
            kegg_organism_code (str): KEGG organism code to prepend to the kegg_id if not part of it already.
                Example: ``eco:b1244``, eco is the organism code
            kegg_seq_file (str): Path to KEGG FASTA file
            kegg_metadata_file (str): Path to KEGG metadata file (raw KEGG format)
            set_as_representative (bool): If this KEGG ID should be set as the representative sequence
            download (bool): If the KEGG sequence and metadata files should be downloaded if not provided
            outdir (str): Where the sequence and metadata files should be downloaded to
            force_rerun (bool): If ID should be reloaded and files redownloaded

        Returns:
            KEGGProp: object contained in the sequences attribute

        """
        if download:
            if not outdir:
                outdir = self.sequence_dir
                if not outdir:
                    raise ValueError('Output directory must be specified')

        if kegg_organism_code:
            kegg_id = kegg_organism_code + ':' + kegg_id

        # If we have already loaded the KEGG ID
        if self.sequences.has_id(kegg_id):
            # Remove it if we want to force rerun things
            if force_rerun:
                existing = self.sequences.get_by_id(kegg_id)
                self.sequences.remove(existing)
            # Otherwise just get that KEGG object
            else:
                log.debug('{}: KEGG ID already present in list of sequences'.format(kegg_id))
                kegg_prop = self.sequences.get_by_id(kegg_id)

        # Check again (instead of else) in case we removed it if force rerun
        if not self.sequences.has_id(kegg_id):
            kegg_prop = KEGGProp(kegg_id, kegg_seq_file, kegg_metadata_file)
            if download:
                kegg_prop.download_seq_file(outdir, force_rerun)
                kegg_prop.download_metadata_file(outdir, force_rerun)

            if kegg_prop.sequence_path:
                # Check if KEGG sequence matches a potentially set representative sequence
                # Do not add any info if a UniProt ID was already mapped though, we want to use that
                if self.representative_sequence:
                    if not self.representative_sequence.uniprot:
                        if kegg_prop.equal_to(self.representative_sequence):
                            # Update the representative sequence field with KEGG metadata
                            self.representative_sequence.update(kegg_prop.get_dict(), only_keys=['sequence_path',
                                                                                                 'metadata_path',
                                                                                                 'kegg',
                                                                                                 'description',
                                                                                                 'taxonomy',
                                                                                                 'id',
                                                                                                 'pdbs',
                                                                                                 'uniprot',
                                                                                                 'seq_record',
                                                                                                 'gene_name',
                                                                                                 'refseq'])
                        else:
                            # TODO: add option to use manual or kegg sequence if things do not match
                            log.warning('{}: representative sequence does not match mapped KEGG sequence.'.format(self.id))

            self.sequences.append(kegg_prop)

        if set_as_representative:
            self._representative_sequence_setter(kegg_prop)

        return self.sequences.get_by_id(kegg_id)

    def load_uniprot(self, uniprot_id, uniprot_seq_file=None, uniprot_metadata_file=None,
                     set_as_representative=False, download=False, outdir=None, force_rerun=False):
        """Load a UniProt ID and associated sequence/metadata files into the sequences attribute.

        Args:
            uniprot_id:
            uniprot_seq_file:
            uniprot_metadata_file:
            set_as_representative:
            download:
            outdir:
            force_rerun:

        Returns:
            UniProtProp:

        """
        if download:
            if not outdir:
                outdir = self.sequence_dir
                if not outdir:
                    raise ValueError('Output directory must be specified')

        # If we have already loaded the KEGG ID
        if self.sequences.has_id(uniprot_id):
            # Remove it if we want to force rerun things
            if force_rerun:
                existing = self.sequences.get_by_id(uniprot_id)
                self.sequences.remove(existing)
            # Otherwise just get that KEGG object
            else:
                log.debug('{}: KEGG ID already present in list of sequences'.format(uniprot_id))
                uniprot_prop = self.sequences.get_by_id(uniprot_id)

        if not self.sequences.has_id(uniprot_id):
            uniprot_prop = UniProtProp(uniprot_id, uniprot_seq_file, uniprot_metadata_file)
            if download:
                uniprot_prop.download_seq_file(outdir, force_rerun)
                uniprot_prop.download_metadata_file(outdir, force_rerun)

            # Also check if UniProt sequence matches a potentially set representative sequence
            if self.representative_sequence:
                # Test equality
                if uniprot_prop.equal_to(self.representative_sequence):
                    # Update the representative sequence field with UniProt metadata
                    self.representative_sequence.update(uniprot_prop.get_dict(), only_keys=['sequence_path',
                                                                                            'metadata_path',
                                                                                            'kegg',
                                                                                            'description',
                                                                                            'taxonomy',
                                                                                            'id',
                                                                                            'pdbs',
                                                                                            'uniprot',
                                                                                            'seq_record',
                                                                                            'gene_name',
                                                                                            'refseq'])
                else:
                    # TODO: add option to use manual or uniprot sequence if things do not match
                    log.warning('{}: representative sequence does not match mapped UniProt sequence'.format(self.id))
            self.sequences.append(uniprot_prop)

        if set_as_representative:
            self._representative_sequence_setter(uniprot_prop)

        return self.sequences.get_by_id(uniprot_id)

    def load_manual_sequence_file(self, ident, seq_file, copy_file=False, outdir=None, set_as_representative=False):
        """Load a manual sequence given as a FASTA file and optionally set it as the representative sequence.
        Also store it in the sequences attribute.

        Args:
            ident:
            seq_file:
            copy_file:
            outdir:
            set_as_representative:

        """
        if copy_file:
            if not outdir:
                outdir = self.sequence_dir
                if not outdir:
                    raise ValueError('Output directory must be specified')
            shutil.copy(seq_file, outdir)
            seq_file = op.join(outdir, seq_file)

        manual_sequence = SeqProp(ident=ident, sequence_path=seq_file)
        self.sequences.append(manual_sequence)

        if set_as_representative:
            self._representative_sequence_setter(manual_sequence)

        return self.sequences.get_by_id(ident)

    def load_manual_sequence(self, seq, ident=None, write_fasta_file=False, outname=None, outdir=None,
                             set_as_representative=False, force_rewrite=False):
        """Load a manual sequence given as a string and optionally set it as the representative sequence.
        Also store it in the sequences attribute.

        Args:
            seq (str, Seq, SeqRecord): Sequence string, Biopython Seq or SeqRecord object
            ident (str): Optional identifier for the sequence, required if seq is a string. Also will override existing
                IDs in Seq or SeqRecord objects if set.
            write_fasta_file:
            outname:
            outdir:
            set_as_representative:
            force_rewrite:

        Returns:

        """
        if write_fasta_file:
            if not outdir:
                outdir = self.sequence_dir
                if not outdir:
                    raise ValueError('Output directory must be specified')

        if isinstance(seq, str) or isinstance(seq, Seq):
            if not ident:
                raise ValueError('ID must be specified if sequence is a string or Seq object')
        else:
            if not ident:
                # Use ID from SeqRecord ID if new ID not provided
                ident = seq.id
            else:
                # Overwrite SeqRecord ID with new ID if provided
                seq.id = ident

        if not outname:
            outname = ident



        manual_sequence = SeqProp(ident=ident, seq=seq, write_fasta_file=write_fasta_file,
                                  outname=outname, outdir=outdir, force_rewrite=force_rewrite)
        self.sequences.append(manual_sequence)

        if set_as_representative:
            self._representative_sequence_setter(manual_sequence)

        return self.sequences.get_by_id(ident)

    def _representative_sequence_setter(self, seq_prop):
        """Make a copy of a SeqProp object and store it as the representative. Only keep certain attributes"""

        self.representative_sequence = SeqProp(ident=seq_prop.id, sequence_path=seq_prop.sequence_path) #,
                                               # seq=seq_prop.seq_record)
        self.representative_sequence.update(seq_prop.get_dict(), only_keys=self.__representative_sequence_attributes)
        if self.representative_sequence.seq_record:
            self.representative_sequence.seq_record.id = self.representative_sequence.id
        log.debug('{}: set as representative sequence'.format(seq_prop.id))

    def set_representative_sequence(self, force_rerun=False):
        """Consolidate sequences that were loaded and set a single representative sequence."""

        if len(self.sequences) == 0:
            log.error('{}: no sequences mapped'.format(self.id))
            return self.representative_sequence

        kegg_mappings = self.filter_sequences(KEGGProp)
        if len(kegg_mappings) > 0:
            kegg_to_use = kegg_mappings[0]
            if len(kegg_mappings) > 1:
                log.warning('{}: multiple KEGG mappings found, using the first entry {}'.format(self.id, kegg_to_use.id))

        uniprot_mappings = self.filter_sequences(UniProtProp)

        # If a representative sequence has already been set, nothing needs to be done
        if self.representative_sequence and not force_rerun:
            log.debug('{}: representative sequence already set'.format(self.id))

        # If there is a KEGG annotation and no UniProt annotations, set KEGG as representative
        elif len(kegg_mappings) > 0 and len(uniprot_mappings) == 0:
            self._representative_sequence_setter(kegg_to_use)
            log.debug('{}: representative sequence set from KEGG ID {}'.format(self.id, kegg_to_use.id))

        # If there are UniProt annotations and no KEGG annotations, set UniProt as representative
        elif len(kegg_mappings) == 0 and len(uniprot_mappings) > 0:
            # If there are multiple uniprots rank them by the sum of reviewed (bool) + num_pdbs
            # This way, UniProts with PDBs get ranked to the top, or if no PDBs, reviewed entries
            u_ranker = []
            for u in uniprot_mappings:
                u_ranker.append((u.id, u.ranking_score()))
            sorted_by_second = sorted(u_ranker, key=lambda tup: tup[1], reverse=True)
            best_u_id = sorted_by_second[0][0]

            best_u = uniprot_mappings.get_by_id(best_u_id)
            self._representative_sequence_setter(best_u)
            log.debug('{}: representative sequence set from UniProt ID {}'.format(self.id, best_u_id))

        # If there are both UniProt and KEGG annotations...
        elif len(kegg_mappings) > 0 and len(uniprot_mappings) > 0:
            # Use KEGG if the mapped UniProt is unique, and it has PDBs
            if kegg_to_use.num_pdbs > 0 and not uniprot_mappings.has_id(kegg_to_use.uniprot):
                self._representative_sequence_setter(kegg_to_use)
                log.debug('{}: representative sequence set from KEGG ID {}'.format(self.id, kegg_to_use.id))
            else:
                # If there are multiple uniprots rank them by the sum of reviewed (bool) + num_pdbs
                u_ranker = []
                for u in uniprot_mappings:
                    u_ranker.append((u.id, u.ranking_score()))
                sorted_by_second = sorted(u_ranker, key=lambda tup: tup[1], reverse=True)
                best_u_id = sorted_by_second[0][0]

                best_u = uniprot_mappings.get_by_id(best_u_id)
                self._representative_sequence_setter(best_u)
                log.debug('{}: Representative sequence set from UniProt ID {}'.format(self.id, best_u_id))

        return self.representative_sequence

    def pairwise_align_sequences_to_representative(self, gapopen=10, gapextend=0.5, outdir=None,
                                                   engine='needle', parse=True, force_rerun=False):
        """Align all sequences in the sequences attribute to the representative sequence.
        Stores the alignments the representative_sequence.sequence_alignments DictList

        Args:
            gapopen:
            gapextend:
            outdir:
            engine:
            parse (bool): Store locations of mutations, insertions, and deletions in the alignment object (as an annotation)
            force_rerun:

        Returns:

        """
        if not outdir:
            outdir = self.sequence_dir
            if not outdir:
                raise ValueError('Output directory must be specified')

        for seq in self.sequences:
            aln_id = '{}_{}'.format(self.id, seq.id)
            outfile = '{}.needle'.format(aln_id)

            if not self.representative_sequence:
                log.error('{}: no representative sequence set, skipping gene'.format(self.id))
                continue

            if self.representative_sequence.sequence_alignments.has_id(aln_id):
                log.debug('{}: alignment already completed'.format(seq.id))
                continue

            if not seq.seq_str:
                log.error('{}: no sequence stored, skipping alignment'.format(seq.id))
                continue

            # Don't need to compare sequence to itself
            if seq.id == self.representative_sequence.id:
                continue

            aln = ssbio.protein.sequence.utils.alignment.pairwise_sequence_alignment(a_seq=self.representative_sequence.seq_str,
                                                                                     a_seq_id=self.id,
                                                                                     b_seq=seq.seq_str,
                                                                                     b_seq_id=seq.id,
                                                                                     gapopen=gapopen, gapextend=gapextend,
                                                                                     engine=engine,
                                                                                     outdir=outdir,
                                                                                     outfile=outfile,
                                                                                     force_rerun=force_rerun)
            # Add an identifier to the MultipleSeqAlignment object for storage in a DictList
            aln.id = aln_id
            aln.annotations['a_seq'] = self.representative_sequence.id
            aln.annotations['b_seq'] = seq.id

            if parse:
                aln_df = ssbio.protein.sequence.utils.alignment.get_alignment_df(a_aln_seq=str(list(aln)[0].seq),
                                                                                 b_aln_seq=str(list(aln)[1].seq))
                aln.annotations['mutations'] = ssbio.protein.sequence.utils.alignment.get_mutations(aln_df)
                aln.annotations['deletions'] = ssbio.protein.sequence.utils.alignment.get_deletions(aln_df)
                aln.annotations['insertions'] = ssbio.protein.sequence.utils.alignment.get_insertions(aln_df)

            self.representative_sequence.sequence_alignments.append(aln)

    def blast_representative_sequence_to_pdb(self, seq_ident_cutoff=0, evalue=0.0001, display_link=False,
                                             outdir=None, force_rerun=False):
        """BLAST repseq to PDB and return the list of new structures added, also saves df_pdb_blast"""
        # Check if a representative sequence was set
        if not self.representative_sequence:
            log.warning('{}: no representative sequence set, cannot BLAST'.format(self.id))
            return None

        # Also need to check if a sequence has been stored
        if not self.representative_sequence.seq_str:
            log.warning('{}: no representative sequence loaded, cannot BLAST'.format(self.id))
            return None

        # BLAST the sequence to the PDB
        blast_results = self.representative_sequence.blast_pdb(seq_ident_cutoff=seq_ident_cutoff,
                                                               evalue=evalue,
                                                               display_link=display_link,
                                                               outdir=outdir,
                                                               force_rerun=force_rerun)

        new_pdbs = []
        # Add BLAST results to the list of structures
        if blast_results:
            # Filter for new BLASTed PDBs
            pdbs = [x['hit_pdb'].lower() for x in blast_results]
            new_pdbs = [y for y in pdbs if not self.structures.has_id(y)]
            if new_pdbs:
                log.debug('{}: adding {} PDBs from BLAST results'.format(self.id, len(new_pdbs)))
            blast_results = [z for z in blast_results if z['hit_pdb'].lower() in new_pdbs]

            for blast_result in blast_results:
                pdb = blast_result['hit_pdb'].lower()
                chains = blast_result['hit_pdb_chains']

                for chain in chains:
                    # load_pdb will append this protein to the list of structures
                    new_pdb = self.load_pdb(pdb_id=pdb, mapped_chains=chain)
                    new_pdb.add_chain_ids(chain)
                    new_chain = new_pdb.chains.get_by_id(chain)

                    # Store BLAST results within the chain
                    new_chain.blast_results = blast_result

        return new_pdbs

    @property
    def df_pdb_blast(self):
        """Get a dataframe of PDB BLAST results"""

        blast_results_pre_df = []

        for p in self.get_experimental_structures():
            for c in p.chains:
                if hasattr(c, 'blast_results'):
                    # Summary dataframe
                    infodict = p.get_dict_with_chain(chain=c.id)['blast_results']
                    infodict['pdb_id'] = p.id
                    infodict['pdb_chain_id'] = c.id
                    blast_results_pre_df.append(infodict)

        cols = ['pdb_id', 'pdb_chain_id', 'hit_score', 'hit_evalue', 'hit_percent_similar',
                'hit_percent_ident', 'hit_percent_gaps', 'hit_num_ident', 'hit_num_similar', 'hit_num_gaps']
        df = pd.DataFrame.from_records(blast_results_pre_df, columns=cols).set_index('pdb_id')
        return ssbio.utils.clean_df(df)

    def map_uniprot_to_pdb(self, seq_ident_cutoff=0.0, outdir=None, force_rerun=False):
        """"""
        if not self.representative_sequence:
            log.error('{}: no representative sequence set, cannot use best structures API'.format(self.id))
            return None

        # Check if a UniProt ID is attached to the representative sequence
        uniprot_id = self.representative_sequence.uniprot
        if not uniprot_id:
            log.error('{}: no representative UniProt ID set, cannot use best structures API'.format(self.id))
            return None

        if not outdir:
            outdir = self.sequence_dir
            if not outdir:
                raise ValueError('Output directory must be specified')

        best_structures = ssbio.databases.pdb.best_structures(uniprot_id,
                                                              outname='{}_best_structures'.format(custom_slugify(uniprot_id)),
                                                              outdir=outdir,
                                                              seq_ident_cutoff=seq_ident_cutoff,
                                                              force_rerun=force_rerun)

        new_pdbs = []
        if best_structures:
            rank = 1
            for best_structure in best_structures:
                currpdb = str(best_structure['pdb_id'].lower())
                new_pdbs.append(currpdb)
                currchain = str(best_structure['chain_id'])

                # load_pdb will append this protein to the list
                new_pdb = self.load_pdb(pdb_id=currpdb, mapped_chains=currchain)

                # Also add this chain to the chains attribute so we can save the
                # info we get from best_structures
                new_pdb.add_chain_ids(currchain)

                pdb_specific_keys = ['experimental_method', 'resolution']
                chain_specific_keys = ['coverage', 'start', 'end', 'unp_start', 'unp_end']

                new_pdb.update(best_structure, only_keys=pdb_specific_keys)
                new_chain = new_pdb.chains.get_by_id(currchain)
                new_chain.update(best_structure, only_keys=chain_specific_keys)
                new_chain.update({'rank': rank})

                rank += 1

            log.debug('{}, {}: {} PDB/chain pairs mapped'.format(self.id, uniprot_id, len(best_structures)))
        else:
            log.debug('{}, {}: no PDB/chain pairs mapped'.format(self.id, uniprot_id))

        return new_pdbs

    @property
    def df_pdb_ranking(self):
        """Get a dataframe of UniProt -> best structure in PDB results"""

        best_structures_pre_df = []

        chain_specific_keys = ['coverage', 'start', 'end', 'unp_start', 'unp_end', 'rank']

        for p in self.get_experimental_structures():
            for c in p.chains:
                if hasattr(c, 'rank'):
                    # Summary dataframe
                    infodict = p.get_dict_with_chain(chain=c.id, df_format=True, chain_keys=chain_specific_keys)
                    infodict['pdb_id'] = p.id
                    infodict['pdb_chain_id'] = c.id
                    infodict['uniprot'] = self.representative_sequence.uniprot
                    best_structures_pre_df.append(infodict)

        cols = ['uniprot', 'pdb_id', 'pdb_chain_id', 'experimental_method', 'resolution', 'coverage',
                'taxonomy_name', 'start', 'end', 'unp_start', 'unp_end', 'rank']
        df = pd.DataFrame.from_records(best_structures_pre_df, columns=cols).set_index(['pdb_id', 'pdb_chain_id'])
        return ssbio.utils.clean_df(df)

    def load_pdb(self, pdb_id, mapped_chains=None, pdb_file=None, file_type=None, set_as_representative=False, force_rerun=False):
        """Load a PDB ID into the structures attribute.

        Args:
            pdb_id (str): PDB ID
            mapped_chains (str, list): Chain ID or list of IDs which you are interested in
            pdb_file (str): Path to PDB file
            file_type (str): Type of PDB file
            set_as_representative (bool): If this structure should be set as the representative structure
            parse (bool): If the structure's 3D coordinates and chains should be parsed

        Returns:
            PDBProp: The object that is now contained in the structures attribute

        """
        pdb_id = pdb_id.lower()

        if self.structures.has_id(pdb_id):
            if force_rerun:
                existing = self.structures.get_by_id(pdb_id)
                self.structures.remove(existing)
            else:
                log.debug('{}: PDB ID already present in list of structures'.format(pdb_id))
                pdb = self.structures.get_by_id(pdb_id)

                if mapped_chains:
                    pdb.add_mapped_chain_ids(mapped_chains)
                if pdb_file:
                    pdb.load_structure_path(pdb_file, file_type)

        if not self.structures.has_id(pdb_id):
            pdb = PDBProp(ident=pdb_id, chains=mapped_chains, structure_path=pdb_file, file_type=file_type)
            if mapped_chains:
                pdb.add_mapped_chain_ids(mapped_chains)
            self.structures.append(pdb)

        if set_as_representative:
            self._representative_structure_setter(pdb)

        return self.structures.get_by_id(pdb_id)

    def load_itasser_folder(self, ident, itasser_folder, organize=False, outdir=None, organize_name=None,
                            set_as_representative=False, create_dfs=False, force_rerun=False):
        """Load the results folder from I-TASSER, copy structure files over, and create summary dataframes.

        Args:
            ident: I-TASSER ID
            itasser_folder: Path to results folder
            organize (bool): If select files from modeling should be copied to the Protein directory
            outdir (str): Path to directory where files will be copied and organized to
            organize_name (str): Basename of files to rename results to. If not provided, will use id attribute.
            set_as_representative: If this structure should be set as the representative structure
            create_dfs: If summary dataframes should be created
            parse: If the structure's 3D coordinates and chains should be parsed
            force_rerun:

        Returns:
            ITASSERProp: StructProp object stored in the structures list.

        """
        if organize:
            if not outdir:
                outdir = self.structure_dir
                if not outdir:
                    raise ValueError('Directory to copy results to must be specified')

        if self.structures.has_id(ident):
            if force_rerun:
                existing = self.structures.get_by_id(ident)
                self.structures.remove(existing)
            else:
                log.debug('{}: already present in list of structures'.format(ident))
                itasser = self.structures.get_by_id(ident)

        if not self.structures.has_id(ident):
            itasser = ITASSERProp(ident, itasser_folder, create_dfs=create_dfs)
            self.structures.append(itasser)

        if set_as_representative:
            self._representative_structure_setter(itasser)

        if organize:
            if itasser.model_file:
                # The name of the actual pdb file will be $GENEID_model1.pdb
                if not organize_name:
                    new_itasser_name = self.id + '_model1'
                else:
                    new_itasser_name = organize_name

                # Additional results will be stored in a subdirectory
                dest_itasser_extra_dir = op.join(outdir, '{}_itasser'.format(new_itasser_name))
                ssbio.utils.make_dir(dest_itasser_extra_dir)

                # Copy the model1.pdb and also create summary dataframes
                itasser.copy_results(copy_to_dir=outdir, rename_model_to=new_itasser_name, force_rerun=force_rerun)
                itasser.save_dataframes(outdir=dest_itasser_extra_dir)

        return self.structures.get_by_id(ident)

    def load_generic_structure(self, ident, structure_file=None, is_experimental=False, set_as_representative=False, force_rerun=False):
        if self.structures.has_id(ident):
            if force_rerun:
                existing = self.structures.get_by_id(ident)
                self.structures.remove(existing)
            else:
                log.warning('{}: already present in list of structures'.format(ident))
                model = self.structures.get_by_id(ident)

        if not self.structures.has_id(ident):
            model = StructProp(ident=ident, structure_path=structure_file,
                               is_experimental=is_experimental)
            self.structures.append(model)

        if set_as_representative:
            self._representative_structure_setter(model)

        return self.structures.get_by_id(ident)

    @property
    def df_homology_models(self):
        """Get a dataframe of homology models"""

        itasser_pre_df = []

        df_cols = ['id', 'model_file', 'model_date', 'difficulty',
                   'top_template_pdb', 'top_template_chain', 'c_score',
                   'tm_score', 'tm_score_err', 'rmsd', 'rmsd_err',
                   'top_bsite_site_num', 'top_bsite_c_score', 'top_bsite_cluster_size', 'top_bsite_algorithm',
                   'top_bsite_pdb_template_id', 'top_bsite_pdb_template_chain', 'top_bsite_pdb_ligand',
                   'top_bsite_binding_location_coords', 'top_bsite_c_score_method', 'top_bsite_binding_residues',
                   'top_bsite_ligand_cluster_counts',
                   'top_ec_pdb_template_id', 'top_ec_pdb_template_chain', 'top_ec_tm_score', 'top_ec_rmsd',
                   'top_ec_seq_ident', 'top_ec_seq_coverage', 'top_ec_c_score', 'top_ec_ec_number',
                   'top_ec_binding_residues',
                   'top_go_mf_go_id', 'top_go_mf_go_term', 'top_go_mf_c_score', 'top_go_bp_go_id', 'top_go_bp_go_term',
                   'top_go_bp_c_score', 'top_go_cc_go_id', 'top_go_cc_go_term', 'top_go_cc_c_score']

        for p in self.get_homology_models():
            # Summary dataframe
            new_homology_dict = p.get_dict(df_format=True, only_keys=df_cols)
            itasser_pre_df.append(new_homology_dict)

        df = pd.DataFrame.from_records(itasser_pre_df, columns=df_cols).set_index('id')
        return ssbio.utils.clean_df(df)

    def pdb_downloader_and_metadata(self, outdir=None, pdb_file_type=None, force_rerun=False):
        """Download experimental structures"""
        if not outdir:
            outdir = self.structure_dir
            if not outdir:
                raise ValueError('Output directory must be specified')

        if not pdb_file_type:
            pdb_file_type = self.pdb_file_type

        # Check if we have any PDBs
        if self.num_structures_experimental == 0:
            log.debug('{}: no structures available - nothing will be downloaded'.format(self.id))
            return

        downloaded_pdb_ids = []
        # Download the PDBs
        for s in self.get_experimental_structures():
            log.debug('{}: downloading structure file from the PDB...'.format(s.id))
            s.download_structure_file(outdir=outdir, file_type=pdb_file_type, force_rerun=force_rerun)
            # Download the mmCIF header file to get additional information
            if 'cif' not in pdb_file_type:
                s.download_cif_header_file(outdir=outdir, force_rerun=force_rerun)
            downloaded_pdb_ids.append(s.id)

        return downloaded_pdb_ids

    @property
    def df_pdb_metadata(self):
        """Get a dataframe of PDB metadata (PDBs have to be downloaded first)"""
        if self.num_structures == 0:
            log.error('No experimental PDB structures have been mapped to protein')
            return pd.DataFrame()

        pdb_pre_df = []

        for p in self.get_experimental_structures():
            if not p.structure_file:
                log.error('{}: PDB file has not been downloaded, run "pdb_downloader_and_metadata" \
                    to download all structures'.format(p.id))
                continue

            infodict = p.get_dict(df_format=True)
            infodict['pdb_id'] = p.id
            pdb_pre_df.append(infodict)

        cols = ['pdb_id', 'pdb_title', 'description', 'experimental_method', 'mapped_chains',
                'resolution', 'chemicals', 'date', 'taxonomy_name',
                'structure_file']
        df = pd.DataFrame.from_records(pdb_pre_df, columns=cols).set_index('pdb_id')
        return ssbio.utils.clean_df(df)

    def _representative_structure_setter(self, struct_prop, keep_chain, new_id=None, clean=True,
                                         out_suffix='_clean', outdir=None):
        """Set the representative structure by 1) cleaning it and 2) copying over attributes of the original structure.

        The structure is copied because the chains stored may change, and cleaning it makes a new PDB file.

        Args:
            struct_prop (StructProp): StructProp object to set as representative
            keep_chain (str, list): List of chains to keep
            new_id (str): New ID to call this structure, for example 1abc-D to represent PDB 1abc, chain D
            clean (bool): If the PDB file should be cleaned (see ssbio.structure.utils.cleanpdb)
            out_suffix (str): Suffix to append to clean PDB file
            outdir (str): Path to output directory

        Returns:
            StructProp: representative structure

        """
        if not outdir:
            outdir = self.structure_dir
            if not outdir:
                raise ValueError('Output directory must be specified')

        if not new_id:
            new_id = struct_prop.id

        # If the structure is to be cleaned, and which chain to keep
        if clean:
            final_pdb = struct_prop.clean_structure(outdir=outdir, out_suffix=out_suffix, keep_chains=keep_chain)
        else:
            final_pdb = struct_prop.structure_path

        self.representative_structure = StructProp(ident=new_id, chains=keep_chain, mapped_chains=keep_chain,
                                                   structure_path=final_pdb, file_type='pdb',
                                                   representative_chain=keep_chain)

        self.representative_structure.update(struct_prop.get_dict_with_chain(chain=keep_chain),
                                             only_keys=self.__representative_structure_attributes,
                                             overwrite=True)

        # Save the original PDB ID as an extra attribute
        if struct_prop.is_experimental:
            self.representative_structure.original_pdb_id = struct_prop.id

        # STORE REPRESENTATIVE CHAIN RESNUMS in the representative sequence seqrecord letter_annotations
        # Get the alignment
        alnid = '{}_{}'.format(self.representative_sequence.id, self.representative_structure.id)
        aln = self.representative_sequence.structure_alignments.get_by_id(alnid)
        # Get the mapping and store it in .seq_record.letter_annotations['repchain_resnums']
        aln_df = ssbio.protein.sequence.utils.alignment.get_alignment_df(aln[0], aln[1])
        repchain_resnums = aln_df[pd.notnull(aln_df.id_a_pos)].id_b_pos.tolist()
        self.representative_sequence.letter_annotations['repchain_resnums'] = repchain_resnums

        # Also need to parse the clean structure and save its sequence..
        self.representative_structure.parse_structure()

    def set_representative_structure(self, seq_outdir=None, struct_outdir=None, pdb_file_type=None,
                                     engine='needle', always_use_homology=False,
                                     seq_ident_cutoff=0.5, allow_missing_on_termini=0.2,
                                     allow_mutants=True, allow_deletions=False,
                                     allow_insertions=False, allow_unresolved=True,
                                     force_rerun=False):
        """Set a representative structure from a structure in self.structures

        Args:
            seq_outdir (str): Path to output directory of sequence alignment files
            struct_outdir (str): Path to output directory of structure files
            pdb_file_type (str): pdb, pdb.gz, mmcif, cif, cif.gz, xml.gz, mmtf, mmtf.gz - PDB structure file type that
                should be downloaded
            engine (str): "needle" or "biopython" - which pairwise sequence alignment engine should be used
                needle is the standard EMBOSS tool to run pairwise alignments
                biopython is Biopython's implementation of needle. Results can differ!
            always_use_homology (bool): If homology models should always be set as the representative structure
            seq_ident_cutoff (float): Percent sequence identity cutoff, in decimal form
            allow_missing_on_termini (float): Percentage of the total length of the reference sequence which will be ignored
                when checking for modifications. Example: if 0.1, and reference sequence is 100 AA, then only residues
                5 to 95 will be checked for modifications.
            allow_mutants (bool): If mutations should be allowed or checked for
            allow_deletions (bool): If deletions should be allowed or checked for
            allow_insertions (bool): If insertions should be allowed or checked for
            allow_unresolved (bool): If unresolved residues should be allowed or checked for
            force_rerun (bool): If sequence to structure alignment should be rerun

        Returns:
            StructProp: Representative structure from the list of structures.

            This is a not a map to the original structure, it is copied from its reference.

        """
        log.debug('{}: setting representative structure'.format(self.id))

        if len(self.structures) == 0:
            log.debug('{}: no structures available'.format(self.id))
            return None

        if not self.representative_sequence:
            log.error('{}: no representative sequence to compare structures to'.format(self.id))
            return None

        if self.representative_structure and not force_rerun:
            log.debug('{}: representative structure already set'.format(self.id))
            return self.representative_structure

        if not pdb_file_type:
            pdb_file_type = self.pdb_file_type

        if not seq_outdir:
            seq_outdir = self.sequence_dir
            if not seq_outdir:
                raise ValueError('Sequence output directory must be specified')

        if not struct_outdir:
            struct_outdir = self.structure_dir
            if not struct_outdir:
                raise ValueError('Structure output directory must be specified')

        has_homology = False
        has_pdb = False
        use_homology = False
        use_pdb = False

        if self.num_structures_homology > 0:
            has_homology = True
        if self.num_structures_experimental > 0:
            has_pdb = True

        # If we mark to always use homology, use it if it exists
        if always_use_homology:
            if has_homology:
                use_homology = True
            elif has_pdb:
                use_pdb = True
        # If we don't always want to use homology, use PDB if it exists
        else:
            if has_homology and has_pdb:
                use_pdb = True
                use_homology = True
            elif has_homology and not has_pdb:
                use_homology = True
            elif has_pdb and not has_homology:
                use_pdb = True

        if use_pdb:
            # Put PDBs through QC/QA
            log.debug('{}: checking quality of experimental structures'.format(self.id))
            all_pdbs = self.get_experimental_structures()

            for pdb in all_pdbs:
                # Download the structure and parse it
                # This will add all chains to the mapped_chains attribute if there are none
                try:
                    pdb.download_structure_file(outdir=struct_outdir, file_type=pdb_file_type, force_rerun=force_rerun)
                    # Download the mmCIF header file to get additional information
                    if 'cif' not in pdb_file_type:
                        pdb.download_cif_header_file(outdir=struct_outdir, force_rerun=force_rerun)
                except (requests.exceptions.HTTPError, URLError):
                    log.error('{}: structure file could not be downloaded in {} format'.format(pdb, pdb_file_type))
                    continue
                    # TODO: add try/except to download cif file as fallback like below?

                # TODO: clean up these try/except things
                try:
                    pdb.align_seqprop_to_mapped_chains(seqprop=self.representative_sequence,
                                                       outdir=seq_outdir,
                                                       engine=engine,
                                                       parse=False, force_rerun=force_rerun)
                except PDBConstructionException:
                    log.error('{}: unable to parse structure file as {}. Falling back to mmCIF format.'.format(pdb, pdb_file_type))
                    # Fall back to using mmCIF file if structure cannot be parsed
                    try:
                        pdb.download_structure_file(outdir=struct_outdir, file_type='cif',
                                                    force_rerun=force_rerun)
                        # Download the mmCIF header file to get additional information
                        if 'cif' not in pdb_file_type:
                            pdb.download_cif_header_file(outdir=struct_outdir, force_rerun=force_rerun)
                    except (requests.exceptions.HTTPError, URLError):
                        log.error('{}: structure file could not be downloaded'.format(pdb))
                        continue
                    pdb.align_seqprop_to_mapped_chains(seqprop=self.representative_sequence,
                                                       outdir=seq_outdir, engine=engine, parse=False,
                                                       force_rerun=force_rerun)

                best_chain = pdb.sequence_quality_checker(reference_seq=self.representative_sequence,
                                                          seq_ident_cutoff=seq_ident_cutoff,
                                                          allow_missing_on_termini=allow_missing_on_termini,
                                                          allow_mutants=allow_mutants, allow_deletions=allow_deletions,
                                                          allow_insertions=allow_insertions, allow_unresolved=allow_unresolved)

                if best_chain:
                    try:
                        self._representative_structure_setter(struct_prop=pdb,
                                                              new_id='{}-{}'.format(pdb.id, best_chain.id),
                                                              clean=True,
                                                              out_suffix='-{}_clean'.format(best_chain.id),
                                                              keep_chain=best_chain.id,
                                                              outdir=struct_outdir)
                    except:
                        # TODO: inspect causes of these errors - most common is Biopython PDBParser error
                        logging.exception("Unknown error with PDB ID {}".format(pdb.id))
                        continue
                    log.debug('{}-{}: set as representative structure'.format(pdb.id, best_chain.id))
                    pdb.reset_chain_seq_records()
                    return self.representative_structure
                else:
                    pdb.reset_chain_seq_records()
            else:
                log.debug('{}: no experimental structures meet cutoffs'.format(self.id))

        # If we are to use homology, save its information in the representative structure field
        if use_homology:
            log.debug('{}: checking quality of homology models'.format(self.id))
            all_models = self.get_homology_models()

            # TODO: homology models are not ordered in any other way other than how they are loaded,
            # rethink this for multiple homology models
            for homology in all_models:
                if not homology.structure_path:
                    log.debug('{}: no homology structure file'.format(self.id))
                    continue

                homology.align_seqprop_to_mapped_chains(seqprop=self.representative_sequence,
                                                        outdir=seq_outdir, engine=engine, parse=False,
                                                        force_rerun=force_rerun)
                best_chain = homology.sequence_quality_checker(reference_seq=self.representative_sequence,
                                                               seq_ident_cutoff=seq_ident_cutoff,
                                                               allow_missing_on_termini=allow_missing_on_termini,
                                                               allow_mutants=allow_mutants,
                                                               allow_deletions=allow_deletions,
                                                               allow_insertions=allow_insertions,
                                                               allow_unresolved=allow_unresolved)

                if best_chain:
                    if not best_chain.id.strip():
                        best_chain_suffix = 'X'
                    else:
                        best_chain_suffix = best_chain.id
                    try:
                        self._representative_structure_setter(struct_prop=homology,
                                                              new_id='{}-{}'.format(homology.id, best_chain_suffix),
                                                              clean=True,
                                                              out_suffix='-{}_clean'.format(best_chain_suffix),
                                                              keep_chain=best_chain.id,
                                                              outdir=struct_outdir)
                    except:
                        # TODO: inspect causes of these errors - most common is Biopython PDBParser error
                        logging.exception("Unknown error with homology model {}".format(homology.id))
                        continue
                    log.debug('{}-{}: set as representative structure'.format(homology.id, best_chain_suffix))
                    homology.reset_chain_seq_records()
                    return self.representative_structure
                else:
                    homology.reset_chain_seq_records()

        log.warning('{}: no structures meet quality checks'.format(self.id))
        return None

    def view_all_mutations(self, grouped=False, color='red', unique_colors=True, structure_opacity=0.5,
                           opacity_range=(0.8,1), scale_range=(1,5), gui=False):
        """Map all sequence alignment mutations to the structure.

        Args:
            grouped (bool): If groups of mutations should be colored and sized together
            color (str): Color of the mutations (overridden if unique_colors=True)
            unique_colors (bool): If each mutation/mutation group should be colored uniquely
            structure_opacity (float): Opacity of the protein structure cartoon representation
            opacity_range (tuple): Min/max opacity values (mutations that show up more will be opaque)
            scale_range (tuple): Min/max size values (mutations that show up more will be bigger)
            gui (bool): If the NGLview GUI should show up

        Returns:
            NGLviewer object

        """
        single, fingerprint = self.representative_sequence.sequence_mutation_summary()

        single_lens = {k: len(v) for k, v in single.items()}
        single_map_to_structure = {}
        for k, v in single_lens.items():
            resnum = int(k[1])
            resnum_to_structure = self.representative_structure.map_repseq_resnums_to_structure_resnums(self.representative_sequence,
                                                                                                        resnum)
            if resnum not in resnum_to_structure:
                log.warning('{}: residue is not available in structure {}'.format(resnum, self.representative_structure.id))
                continue
            new_key = resnum_to_structure[resnum][1]
            single_map_to_structure[new_key] = v

        if not grouped:
            view = self.representative_structure.view_structure_and_highlight_residues_scaled(single_map_to_structure,
                                                                                       color=color, unique_colors=unique_colors,
                                                                                       structure_opacity=structure_opacity,
                                                                                       opacity_range=opacity_range,
                                                                                       scale_range=scale_range,
                                                                                       gui=gui)
            return view

        else:
            fingerprint_lens = {k: len(v) for k, v in fingerprint.items()}
            fingerprint_map_to_structure = {}
            for k, v in fingerprint_lens.items():
                k_list = [int(x[1]) for x in k]
                resnums_to_structure = self.representative_structure.map_repseq_resnums_to_structure_resnums(self.representative_sequence,
                                                                                                             k_list)
                new_key = tuple(y[1] for y in resnums_to_structure.values())
                fingerprint_map_to_structure[new_key] = v

            view = self.representative_structure.view_structure_and_highlight_residues_scaled(fingerprint_map_to_structure,
                                                                                       color=color, unique_colors=unique_colors,
                                                                                       opacity_range=opacity_range,
                                                                                       scale_range=scale_range,
                                                                                       gui=gui)
            return view

    def summarize_protein(self):
        """Gather all possible attributes in the sequences and structures and summarize everything.

        Returns:
            dict:

        """
        d = OrderedDict()
        repseq = self.representative_sequence
        if not self.representative_structure:
            repstruct = StructProp(self.id)
            repchain = repstruct.representative_chain
        else:
            repstruct = self.representative_structure
            repchain = self.representative_structure.representative_chain
        single, fingerprint = self.representative_sequence.sequence_mutation_summary()
        numstrains = len(self.sequences) - 1

        d['Gene ID'] = self.id
        d['Number of sequences'] = len(self.sequences)
        d['Number of structures (total)'] = self.num_structures
        d['Number of structures (experimental)'] = self.num_structures_experimental
        d['Number of structures (homology models)'] = self.num_structures_homology

        # d['------REPRESENTATIVE SEQUENCE PROPERTIES------')
        #     d['Sequence ID'] = repseq.id
        d['Sequence length'] = repseq.sequence_len
        d['Predicted number of transmembrane helices'] = repseq.seq_record.annotations['num_tm_helix-tmhmm']

        # d['------REPRESENTATIVE STRUCTURE PROPERTIES------')
        d['Structure ID'] = repstruct.id
        #     d['Structure representative chain'] = format(repchain.id)))
        d['Structure is experimental'] = repstruct.is_experimental
        # d['Structure origin'] = repstruct.taxonomy_name))
        # d['Structure description'] = repstruct.description))
        if repstruct.reference_seq_top_coverage:
            d['Structure coverage of sequence'] = str(repstruct.reference_seq_top_coverage) + '%'

        # d['------ALIGNMENTS SUMMARY------')
        #     d['Number of sequence alignments'] = len(repseq.sequence_alignments)))
        #     d['Number of structure alignments'] = len(repseq.structure_alignments)))

        singles = []
        for k, v in single.items():
            k = [str(x) for x in k]
            if len(v) / numstrains >= 0.01:
                singles.append(''.join(k))  # len(v) is the number of strains
        d['Mutations that show up in more than 10% of strains'] = ';'.join(singles)

        allfingerprints = []
        for k, v in fingerprint.items():
            if len(v) / numstrains >= 0.01:
                fingerprints = []
                for m in k:
                    y = [str(x) for x in m]
                    fingerprints.append(''.join(y))
                allfingerprints.append('-'.join(fingerprints))
        d['Mutation groups that show up in more than 10% of strains'] = ';'.join(allfingerprints)

        return d

    def __json_decode__(self, **attrs):
        for k, v in attrs.items():
            if k == 'sequences' or k == 'structures':
                setattr(self, k, DictList(v))
            else:
                setattr(self, k, v)


import io
import logging
import os.path as op
from collections import defaultdict

from bioservices import KEGG
from slugify import Slugify

import ssbio.utils
from ssbio.protein.sequence.seqprop import SeqProp

log = logging.getLogger(__name__)
custom_slugify = Slugify(safe_chars='-_')
bs_kegg = KEGG()


class KEGGProp(SeqProp):
    def __init__(self, kegg_id, sequence_path=None, metadata_path=None):

        SeqProp.__init__(self, ident=kegg_id, sequence_path=sequence_path, metadata_path=metadata_path)

        if kegg_id:
            self.kegg = kegg_id

        if metadata_path:
            self.load_metadata_path(metadata_path)

    def load_metadata_path(self, metadata_path):
        SeqProp.load_metadata_path(self, metadata_path)
        self.update(parse_kegg_gene_metadata(metadata_path), overwrite=True)

    def download_seq_file(self, outdir, force_rerun=False):
        kegg_seq_file = download_kegg_aa_seq(gene_id=self.id,
                                             outdir=outdir,
                                             force_rerun=force_rerun)
        if kegg_seq_file:
            self.load_sequence_path(kegg_seq_file)
        else:
            log.warning('{}: no sequence file available'.format(self.id))

    def download_metadata_file(self, outdir, force_rerun=False):
        kegg_metadata_file = download_kegg_gene_metadata(gene_id=self.id,
                                                         outdir=outdir,
                                                         force_rerun=force_rerun)
        if kegg_metadata_file:
            self.load_metadata_path(kegg_metadata_file)
        else:
            log.warning('{}: no metadata file available'.format(self.id))


def download_kegg_gene_metadata(gene_id, outdir=None, force_rerun=False):
    """Download the KEGG flatfile for a KEGG ID and return the path.

    Args:
        gene_id: KEGG gene ID (with organism code), i.e. "eco:1244"
        outdir: optional output directory of metadata

    Returns:
        Path to metadata file

    """
    if not outdir:
        outdir = ''

    # Replace colon with dash in the KEGG gene ID
    outfile = op.join(outdir, '{}.kegg'.format(custom_slugify(gene_id)))

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        raw_text = bs_kegg.get("{}".format(gene_id))
        if raw_text == 404:
            return

        with io.open(outfile, mode='wt', encoding='utf-8') as f:
            f.write(raw_text)

        log.debug('{}: downloaded KEGG metadata file'.format(outfile))
    else:
        log.debug('{}: KEGG metadata file already exists'.format(outfile))

    return outfile


def parse_kegg_gene_metadata(infile):
    """Parse the KEGG flatfile and return a dictionary of metadata.

    Dictionary keys are:
        refseq
        uniprot
        pdbs
        taxonomy

    Args:
        infile: Path to KEGG flatfile

    Returns:
        dict: Dictionary of metadata

    """
    metadata = defaultdict(str)

    with open(infile) as mf:
        kegg_parsed = bs_kegg.parse(mf.read())

    # TODO: additional fields can be parsed

    if 'DBLINKS' in kegg_parsed.keys():
        if 'UniProt' in kegg_parsed['DBLINKS']:
            unis = str(kegg_parsed['DBLINKS']['UniProt']).split(' ')
            # TODO: losing other uniprot ids by doing this
            if isinstance(unis, list):
                metadata['uniprot'] = unis[0]
            else:
                metadata['uniprot'] = unis
        if 'NCBI-ProteinID' in kegg_parsed['DBLINKS']:
            metadata['refseq'] = str(kegg_parsed['DBLINKS']['NCBI-ProteinID'])
    if 'STRUCTURE' in kegg_parsed.keys():
        metadata['pdbs'] = str(kegg_parsed['STRUCTURE']['PDB']).split(' ')
    else:
        metadata['pdbs'] = None
    if 'ORGANISM' in kegg_parsed.keys():
        metadata['taxonomy'] = str(kegg_parsed['ORGANISM'])

    return metadata


def download_kegg_aa_seq(gene_id, outdir=None, force_rerun=False):
    """Download a FASTA sequence of a protein from the KEGG database and return the path.

    Args:
        gene_id: the gene identifier
        outdir: optional path to output directory

    Returns:
        Path to FASTA file

    """
    if not outdir:
        outdir = ''

    outfile = op.join(outdir, '{}.faa'.format(custom_slugify(gene_id)))

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        raw_text = bs_kegg.get("{}".format(gene_id), option='aaseq')
        if raw_text == 404:
            return

        with io.open(outfile, mode='wt', encoding='utf-8') as f:
           f.write(raw_text)

        log.debug('{}: downloaded KEGG FASTA file'.format(outfile))
    else:
        log.debug('{}: KEGG FASTA file already exists'.format(outfile))

    return outfile


def map_kegg_all_genes(organism_code, target_db):
    """Map all of an organism's gene IDs to the target database.

    This is faster than supplying a specific list of genes to map,
    plus there seems to be a limit on the number you can map with a manual REST query anyway.

    Args:
        organism_code: the three letter KEGG code of your organism
        target_db: ncbi-proteinid | ncbi-geneid | uniprot

    Returns:
        Dictionary of ID mapping

    """
    mapping = bs_kegg.conv(target_db, organism_code)

    # strip the organism code from the keys and the identifier in the values
    new_mapping = {}
    for k,v in mapping.items():
        new_mapping[k.replace(organism_code + ':', '')] = str(v.split(':')[1])

    return new_mapping


try:
    from urllib.request import urlopen
    from urllib.request import build_opener
    from urllib.request import HTTPCookieProcessor
    from urllib.parse import urlparse
    from urllib.parse import urlencode
except ImportError:
    from urlparse import urlparse
    from urllib2 import urlopen
    from urllib import urlencode
    from urllib2 import build_opener
    from urllib2 import HTTPCookieProcessor
# TODO: replace urllib usage with six library
import scipy.constants
import math
# import cachetools

# author: Ke Chen

r_j = scipy.constants.R
r_cal = scipy.constants.R / scipy.constants.calorie


def get_folding_rate_for_seq(seq, opn, temp, refT=37):
    """Predict the kinetic folding rate of a protein at temperature T

    Args:
        seq:  protein sequence
        opn:  structural class: all-alpha, all-beta, mixed, unknown
        temp:    temperature
        refT: reference temperature default to 37C

    Returns:
        kinetic folding rate kf (kegg-1)

    """

    slope = 22000  # not many data_dir on this value, it's effect on growth rate very small

    # get folding rate for the reference temperature
    ref_rate = get_folding_rate_from_url(seq, opn)
    preFactor = float(ref_rate) + slope / (float(refT) + 273.15)

    # calculate folding rate at desired temperature
    rate = math.exp(preFactor - slope / (float(temp) + 273.15))

    return rate


# @cachetools.func.ttl_cache(maxsize=500)
def get_folding_rate_from_url(seq, opn):
    """Predict the kinetic folding rate of a protein

    for more information, see http://www.iitm.ac.in/bioinfo/fold-rate/
    it might be better to get sequence directly from ME2.0 (i.e. whatever that's being modeled)
    not from structure here (which may include some weird one letter codes, which need to be fixed)?

    Args:
        seq: protein sequence
        opn: structural class: all-alpha, all-beta, mixed, unknown

    Returns:
        kinetic folding rate ln(kf) ln(kegg-1)

    """

    url = 'http://www.iitm.ac.in/bioinfo/cgi-bin/fold-rate/foldrateCalculator.pl'

    values = {'sequence': seq, 'eqn': opn}
    data = urlencode(values)
    data = data.encode('ASCII')
    response = urlopen(url, data)

    result = str(response.read())
    ind = str.find(result, 'The folding rate,')
    result2 = result[ind:ind + 70]
    ind1 = str.find(result2, '=')
    ind2 = str.find(result2, '/sec')
    rate = result2[ind1 + 2:ind2]
    return rate






import os
import subprocess
import os.path as op
from ssbio import utils
date = utils.Date()

import logging
log = logging.getLogger(__name__)

import pandas as pd


def run_makeblastdb(infile, dbtype, outdir=''):
    """Make the BLAST database for a genome file.

    Args:
        infile (str): path to genome FASTA file
        dbtype (str): "nucl" or "prot" - what format your genome files are in
        outdir (str): path to directory to output database files (default is original folder)

    Returns:
        Paths to BLAST databases.

    """

    # TODO: rewrite using utils function command

    # Output location
    og_dir, name, ext = utils.split_folder_and_path(infile)
    if not outdir:
        outdir = og_dir
    outfile_basename = op.join(outdir, name)

    # Check if BLAST DB was already made
    if dbtype == 'nucl':
        outext = ['.nhr', '.nin', '.nsq']
    elif dbtype == 'prot':
        outext = ['.phr', '.pin', '.psq']
    else:
        raise ValueError('dbtype must be "nucl" or "prot"')
    outfile_all = [outfile_basename + x for x in outext]
    db_made = True
    for f in outfile_all:
        if not op.exists(f):
            db_made = False

    # Run makeblastdb if DB does not exist
    if db_made:
        log.debug('BLAST database already exists at {}'.format(outfile_basename))
        return outfile_all
    else:
        retval = subprocess.call('makeblastdb -in {} -dbtype {} -out {}'.format(infile, dbtype, outfile_basename), shell=True)
        if retval == 0:
            log.debug('Made BLAST database at {}'.format(outfile_basename))
            return outfile_all
        else:
            log.error('Error running makeblastdb, exit code {}'.format(retval))


def run_bidirectional_blast(reference, other_genome, dbtype, outdir=''):
    """BLAST a genome against another, and vice versa.

    This function requires BLAST to be installed, do so by running:
    sudo apt install ncbi-blast+

    Args:
        reference (str): path to "reference" genome, aka your "base strain"
        other_genome (str): path to other genome which will be BLASTed to the reference
        dbtype (str): "nucl" or "prot" - what format your genome files are in
        outdir (str): path to folder where BLAST outputs should be placed

    Returns:
        Paths to BLAST output files.
        (reference_vs_othergenome.out, othergenome_vs_reference.out)

    """

    if dbtype == 'nucl':
        command = 'blastn'
    elif dbtype == 'prot':
        command = 'blastp'
    else:
        raise ValueError('dbtype must be "nucl" or "prot"')

    r_folder, r_name, r_ext = utils.split_folder_and_path(reference)
    g_folder, g_name, g_ext = utils.split_folder_and_path(other_genome)

    # make sure BLAST DBs have been made
    run_makeblastdb(infile=reference, dbtype=dbtype, outdir=r_folder)
    run_makeblastdb(infile=other_genome, dbtype=dbtype, outdir=g_folder)

    # Reference vs genome
    r_vs_g = r_name + '_vs_' + g_name + '_blast.out'
    r_vs_g = op.join(outdir, r_vs_g)
    if op.exists(r_vs_g) and os.stat(r_vs_g).st_size != 0:
        log.debug('{} vs {} BLAST already run'.format(r_name, g_name))
    else:
        cmd = '{} -query {} -db {} -outfmt 6 -out {}'.format(command, reference, op.join(g_folder, g_name), r_vs_g)
        log.debug('Running: {}'.format(cmd))
        retval = subprocess.call(cmd, shell=True)
        if retval == 0:
            log.debug('BLASTed {} vs {}'.format(g_name, r_name))
        else:
            log.error('Error running {}, exit code {}'.format(command, retval))

    # Genome vs reference
    g_vs_r = g_name + '_vs_' + r_name + '_blast.out'
    g_vs_r = op.join(outdir, g_vs_r)
    if op.exists(g_vs_r) and os.stat(g_vs_r).st_size != 0:
        log.debug('{} vs {} BLAST already run'.format(g_name, r_name))
    else:
        cmd = '{} -query {} -db {} -outfmt 6 -out {}'.format(command, other_genome, op.join(r_folder, r_name), g_vs_r)
        log.debug('Running: {}'.format(cmd))
        retval = subprocess.call(cmd, shell=True)
        if retval == 0:
            log.debug('BLASTed {} vs {}'.format(g_name, r_name))
        else:
            log.error('Error running {}, exit code {}'.format(command, retval))

    return r_vs_g, g_vs_r


def calculate_bbh(blast_results_1, blast_results_2, r_name=None, g_name=None, outdir=''):
    """Calculate the best bidirectional BLAST hits (BBH) and save a dataframe of results.

    Args:
        blast_results_1 (str): BLAST results for reference vs. other genome
        blast_results_2 (str): BLAST results for other vs. reference genome
        r_name: Name of reference genome
        g_name: Name of other genome
        outdir: Directory where BLAST results are stored.

    Returns:
        Path to Pandas DataFrame of the BBH results.

    """

    cols = ['gene', 'subject', 'PID', 'alnLength', 'mismatchCount', 'gapOpenCount', 'queryStart', 'queryEnd',
            'subjectStart', 'subjectEnd', 'eVal', 'bitScore']

    if not r_name and not g_name:
        r_name = op.basename(blast_results_1).split('_vs_')[0]
        g_name = op.basename(blast_results_1).split('_vs_')[1].replace('_blast.out', '')

        r_name2 = op.basename(blast_results_2).split('_vs_')[1].replace('_blast.out', '')
        if r_name != r_name2:
            log.warning('{} != {}'.format(r_name, r_name2))

    bbh1 = pd.read_csv(blast_results_1, sep='\t', names=cols)
    bbh2 = pd.read_csv(blast_results_2, sep='\t', names=cols)

    outfile = op.join(outdir, '{}_vs_{}_bbh.csv'.format(r_name, g_name))
    if op.exists(outfile) and os.stat(outfile).st_size != 0:
        log.debug('{} vs {} BLAST BBHs already found at {}'.format(r_name, g_name, outfile))
        return outfile

    out = pd.DataFrame()
    log.debug('Finding BBHs for {} vs. {}'.format(r_name, g_name))

    for g in bbh1[pd.notnull(bbh1.gene)].gene.unique():
        res = bbh1[bbh1.gene == g]
        if len(res) == 0:
            continue
        best_hit = res.ix[res.PID.idxmax()].copy()
        best_gene = best_hit.subject
        res2 = bbh2[bbh2.gene == best_gene]
        if len(res2) == 0:
            continue
        best_hit2 = res2.ix[res2.PID.idxmax()]
        best_gene2 = best_hit2.subject
        if g == best_gene2:
            best_hit['BBH'] = '<=>'
        else:
            best_hit['BBH'] = '->'
        out = pd.concat([out, pd.DataFrame(best_hit).transpose()])

    out.to_csv(outfile)
    log.debug('{} vs {} BLAST BBHs saved at {}'.format(r_name, g_name, outfile))
    return outfile


def create_orthology_matrix(r_name, genome_to_bbh_files, pid_cutoff=None, bitscore_cutoff=None, evalue_cutoff=None,
                            filter_condition='OR', outname='', outdir=''):
    """Create an orthology matrix using best bidirectional BLAST hits (BBH) outputs.

    Args:
        r_name (str): Name of the reference genome
        genome_to_bbh_files (dict): Mapping of genome names to the BBH csv output from the
            :func:`~ssbio.sequence.utils.blast.calculate_bbh` method
        pid_cutoff (float): Minimum percent identity between BLAST hits to filter for in the range [0, 100]
        bitscore_cutoff (float): Minimum bitscore allowed between BLAST hits
        evalue_cutoff (float): Maximum E-value allowed between BLAST hits
        filter_condition (str): 'OR' or 'AND', how to combine cutoff filters. 'OR' gives more results since it
            is less stringent, as you will be filtering for hits with (>80% PID or >30 bitscore or <0.0001 evalue).
        outname: Name of output file of orthology matrix
        outdir: Path to output directory

    Returns:
        str: Path to orthologous genes matrix.

    """
    if outname:
        outfile = op.join(outdir, outname)
    else:
        outfile = op.join(outdir, '{}_orthology.csv'.format(r_name))

    if not pid_cutoff and not bitscore_cutoff and not evalue_cutoff:
        log.warning('No cutoffs supplied, insignificant hits may be reported')

    if filter_condition not in ['OR', 'AND']:
        raise ValueError('Invalid filter condition supplied, must be either "OR" or "AND"')

    if not pid_cutoff:
        pid_cutoff = 0
    if not bitscore_cutoff:
        bitscore_cutoff = 0
    if not evalue_cutoff:
        evalue_cutoff = float('Inf')

    out = pd.DataFrame()
    for g_name, bbh_path in genome_to_bbh_files.items():

        df_bbh = pd.read_csv(bbh_path, index_col=0)
        bidirectional = df_bbh[df_bbh.BBH == '<=>']

        if filter_condition == 'OR':
            data = bidirectional[(bidirectional.PID > pid_cutoff) | (bidirectional.eVal < evalue_cutoff) | (bidirectional.bitScore > bitscore_cutoff)]
        elif filter_condition == 'AND':
            data = bidirectional[(bidirectional.PID > pid_cutoff) & (bidirectional.eVal < evalue_cutoff) & (bidirectional.bitScore > bitscore_cutoff)]

        data.index = data.gene
        data2 = data[['subject']]
        if out.empty:
            out = data2
            out = out.rename(columns={'subject': g_name})
        else:
            out = pd.merge(out, data2, left_index=True, right_index=True, how='outer')
            out = out.rename(columns={'subject': g_name})

    out.to_csv(outfile)
    log.debug('{} orthologous genes saved at {}'.format(r_name, outfile))
    return outfile



import shutil
import os.path as op
import unittest
import cobra
import pandas as pd
import six
from ssbio.pipeline.gempro import GEMPRO
from ssbio.core.genepro import GenePro
from ssbio.utils import Date
from ssbio.databases.kegg import KEGGProp

date = Date()

class TestGEMPRO(unittest.TestCase):
    """Unit tests for the GEMPRO pipeline
    """

    @classmethod
    def setUpClass(self):
        self.GEM_NAME = 'ecoli_test'
        self.ROOT_DIR = op.join('test_files', 'out')
        gem_file = op.join('test_files', 'Ec_core_flux1.xml')

        self.my_gempro = GEMPRO(gem_name=self.GEM_NAME, root_dir=self.ROOT_DIR, gem_file_path=gem_file, gem_file_type='sbml')
        self.base_dir = self.my_gempro.base_dir

        # Remove some genes from this model so testing doesn't take too long
        remove_these = ['b0118', 'b4025', 'b4153', 'b2925', 'b3919', 'b3738', 'b3732', 'b0726', 'b1602', 'b1101',
                        'b3236', 'b0728', 'b1603', 'b2926', 'b0432', 'b3735', 'b0474', 'b4090', 'b3731', 'b0767',
                        'b3737', 'b0724', 'b0008', 'b3403', 'b1779', 'b0727', 'b3956', 'b1676', 'b2935', 'b4232',
                        'b0430', 'b2914', 'b0723', 'b0722', 'b0729', 'b0720', 'b0116', 'b4152', 'b2415', 'b0429',
                        'b0621', 'b0904', 'b1241', 'b1276', 'b1380', 'b1416', 'b1479', 'b1612', 'b1702', 'b2417',
                        'b1852', 'b1854', 'b2276', 'b2277', 'b2278', 'b2279', 'b2280', 'b2281', 'b2282', 'b2283',
                        'b2284', 'b2285', 'b2286', 'b2287', 'b2288', 'b2297', 'b2463', 'b2465', 'b2587', 'b2975',
                        'b3117', 'b3386', 'b3528', 'b3603', 'b3739', 'b3951', 'b3952', 'b4122', 'b4395', 'b0721',
                        'b2029', 'b1136', 'b4015', 'b4014', 'b2976', 'b0114', 'b0115', 'b3916', 'b1723', 'b0755'
                        ]
        for x in remove_these:
            self.my_gempro.genes.remove(self.my_gempro.genes.get_by_id(x))

        # Test the manual adding of any gene ID
        add_these = ['b0002', 'b0003', 'b0004', 'b2092']
        self.my_gempro.add_genes_by_id(add_these)

        self.kegg_will_not_map = ['b1417', 'b2092']
        self.uniprot_will_not_map = ['b1417', 'b2092']

        print(self.base_dir)

    def test_add_genes_by_id(self):
        add_these = ['b0002', 'b0003', 'b0004', 'b2092']
        for x in add_these:
            self.assertTrue(self.my_gempro.genes.has_id(x))

    def test_gene_content(self):
        # Test that genes were correctly removed from self.genes
        remaining = ['b2296', 'b3734', 'b3733', 'b3736', 'b0431', 'b2779', 'b4151', 'b4154', 'b1611',  'b2416',
                     'b0002', 'b0003', 'b0004', 'b1417', 'b2092']
        six.assertCountEqual(self, [x.id for x in self.my_gempro.genes], remaining)

    def test_prep_folders(self):
        # Test if folders were made
        self.assertTrue(op.exists(op.join(self.ROOT_DIR, self.GEM_NAME)))
        folders = ['data', 'structures', 'sequences']
        for f in folders:
            self.assertTrue(op.exists(op.join(self.ROOT_DIR, self.GEM_NAME, f)))

    def test_load_model(self):
        # Test that we can access the model using .model
        self.assertIsInstance(self.my_gempro.model, cobra.Model)

        # Test original model attributes
        self.assertEqual(len(self.my_gempro.model.reactions), 77)
        self.assertEqual(len(self.my_gempro.model.metabolites), 63)
        # genes should be removed from model
        self.assertEqual(len(self.my_gempro.genes), 15)

    def test_gene_info(self):
        # Test that self.genes is a DictList and contains GenePro objects
        self.assertIsInstance(self.my_gempro.genes, cobra.DictList)
        for g in self.my_gempro.genes:
            self.assertIsInstance(g, GenePro)

    def test_kegg_mapping_and_metadata(self):
        """Test that KEGG mapping did these things:
        1. Create the df_kegg_metadata attribute
        2. Save individual sequence and metadata files per gene in the sequence_files folder
        3. Save KEGG related info into each GeneInfo object
        """
        kegg_organism_code = 'eco'
        self.my_gempro.kegg_mapping_and_metadata(kegg_organism_code=kegg_organism_code)
        # print(self.my_gempro.df_kegg_metadata)

        # 1.
        # Test if attribute is a DataFrame and the number of entries in it is equal to the number of genes
        self.assertIsInstance(self.my_gempro.df_kegg_metadata, pd.DataFrame)
        self.assertEqual(len(self.my_gempro.df_kegg_metadata.gene.unique()), len(self.my_gempro.genes))

        ## Tests per gene
        for gene in self.my_gempro.genes:
            g = gene.id

            # 2.
            # Test that gene metadata was downloaded
            # If the gene did not map to a KEGG ID, a folder will still be made for it
            # It will still be in the KEGG DF, just with empty entries
            # It will not have anything downloaded for it
            gene_structure_folder = op.join(self.my_gempro.sequence_dir, g)
            self.assertTrue(op.exists(gene_structure_folder))
            self.assertTrue(g in self.my_gempro.df_kegg_metadata.gene.tolist())

            if g in self.kegg_will_not_map:
                self.assertTrue(g in self.my_gempro.missing_kegg_mapping)
                self.assertFalse(op.exists(op.join(gene_structure_folder, '{}-{}.faa'.format(kegg_organism_code, g))))
            else:
                self.assertTrue(op.exists(op.join(gene_structure_folder, '{}-{}.faa'.format(kegg_organism_code, g))))
                self.assertTrue(op.exists(op.join(gene_structure_folder, '{}-{}.kegg'.format(kegg_organism_code, g))))
                # 3.
                # Test that the sequence_properties attribute has KEGG information in it
                keggs = gene.protein.filter_sequences(KEGGProp)
                for k in keggs:
                    self.assertTrue(k.seq_len > 0)

    def test_uniprot_mapping_and_metadata(self):
        """Test that UniProt mapping did these things:
        1. Create the df_uniprot_metadata attribute
        2. Save individual sequence and metadata files per gene in the sequence_files folder
        3. Save UniProt related info into each GeneInfo object
        """
        self.my_gempro.uniprot_mapping_and_metadata(model_gene_source='ENSEMBLGENOME_ID')
        # print(self.my_gempro.df_uniprot_metadata)

        # 1.
        # Test if attribute is a DataFrame and the number of entries in it is equal to the number of genes
        self.assertIsInstance(self.my_gempro.df_uniprot_metadata, pd.DataFrame)
        self.assertEqual(len(self.my_gempro.df_uniprot_metadata.gene.unique()), len(self.my_gempro.genes))

        ## Tests per gene
        for gene in self.my_gempro.genes:
            g = gene.id

            # 2.
            # Test that gene metadata was downloaded
            # If the gene did not map to a UniProt ID, a folder will still be made for it
            # It will still be in the UniProt DF, just with empty entries
            # It will not have anything downloaded for it
            gene_structure_folder = op.join(self.my_gempro.sequence_dir, g)
            self.assertTrue(op.exists(gene_structure_folder))
            self.assertTrue(g in self.my_gempro.df_uniprot_metadata.gene.tolist())

            if g in self.uniprot_will_not_map:
                self.assertTrue(g in self.my_gempro.missing_uniprot_mapping)
                self.assertFalse(op.exists(op.join(gene_structure_folder, '{}.fasta'.format(u))))
                self.assertFalse(op.exists(op.join(gene_structure_folder, '{}.txt'.format(u))))
            else:
                uniprots = self.my_gempro.df_uniprot_metadata[self.my_gempro.df_uniprot_metadata.gene == g].uniprot.unique().tolist()
                for u in uniprots:
                    self.assertTrue(op.exists(op.join(gene_structure_folder, '{}.fasta'.format(u))))
                    self.assertTrue(op.exists(op.join(gene_structure_folder, '{}.txt'.format(u))))
                    # 3.
                    # Test that the.annotation['sequence'] attribute has UniProt information in it
                    u_prop = gene.protein.sequences.get_by_id(u)
                    self.assertTrue(u_prop.seq_len > 0)

    def test_x_map_uniprot_to_pdb(self):
        self.my_gempro.uniprot_mapping_and_metadata(model_gene_source='ENSEMBLGENOME_ID')
        self.my_gempro.set_representative_sequence()
        self.my_gempro.map_uniprot_to_pdb(seq_ident_cutoff=.9)

        look_at_this_gene = self.my_gempro.genes.get_by_id('b3734').protein
        self.assertTrue(look_at_this_gene.structures.has_id('3oaa'))
        look_at_this_structure = look_at_this_gene.structures.get_by_id('3oaa')
        # TODO: check parsing of taxonomy name
        # self.assertEqual(536056, look_at_this_structure.taxonomy_name)
        look_at_this_chain = look_at_this_structure.chains.get_by_id('A')
        self.assertEqual(513, look_at_this_chain.unp_end)
        self.assertEqual(1, look_at_this_chain.coverage)

    def test_x_blast_pdb(self):
        self.my_gempro.uniprot_mapping_and_metadata(model_gene_source='ENSEMBLGENOME_ID')
        self.my_gempro.set_representative_sequence()
        self.my_gempro.blast_seqs_to_pdb(seq_ident_cutoff=.8, all_genes=True)

        look_at_this_gene = self.my_gempro.genes.get_by_id('b2296').protein
        self.assertTrue(look_at_this_gene.structures.has_id('3slc'))
        look_at_this_structure = look_at_this_gene.structures.get_by_id('3slc')
        look_at_this_chain = look_at_this_structure.chains.get_by_id('A')
        self.assertEqual(0.9425, look_at_this_chain.hit_percent_ident)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree(self.base_dir)


import argparse
import logging

import pandas as pd
from Bio import PDB
from tqdm import tqdm

import ssbio.utils
from ssbio.protein.structure.utils.structureio import StructureIO

log = logging.getLogger(__name__)


def get_msms_df(model, pdb_file, outfile=None, outdir=None, outext='_msms.df', force_rerun=False):
    """Run MSMS (using Biopython) on a Biopython Structure Model and the path to the actual PDB file.

    Returns a dictionary of:
        {chain_id: {resnum1_id: (res_depth, ca_depth)},
                   {resnum2_id: (res_depth, ca_depth)} }
    Depths are in units Angstroms. 1A = 10^-10 m = 1nm

    Args:
        model: Biopython Structure Model
        pdb_file: Path to PDB file

    Returns:
        Pandas DataFrame: ResidueDepth property_dict, reformatted

    """
    # Create the output file name
    outfile = ssbio.utils.outfile_maker(inname=pdb_file, outname=outfile, outdir=outdir, outext=outext)

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        # Run MSMS with Biopython
        try:
            rd = PDB.ResidueDepth(model, pdb_file)
        except AssertionError:
            log.error('{}: unable to run MSMS'.format(pdb_file))
            return pd.DataFrame()

        # Reorganize the results into a csv file
        appender = []
        for k in rd.property_keys:
            x = rd.property_dict[k]
            chain = k[0]
            residue = k[1]
            het = residue[0]
            resnum = residue[1]
            icode = residue[2]
            resdepth = x[0]
            cadepth = x[1]
            appender.append((chain, resnum, icode, resdepth, cadepth))

        df = pd.DataFrame.from_records(appender, columns=['chain', 'resnum', 'icode', 'res_depth', 'ca_depth'])
        df.to_csv(outfile)
    else:
        log.debug('{}: already ran MSMS and force_rerun={}, loading results'.format(outfile, force_rerun))
        df = pd.read_csv(outfile, index_col=0)

    return df


def get_msms_df_on_file(pdb_file, outfile=None, outdir=None, outext='_msms.df', force_rerun=False):
    """Run MSMS (using Biopython) on a PDB file.

    Saves a CSV file of:
        chain: chain ID
        resnum: residue number (PDB numbering)
        icode: residue insertion code
        res_depth: average depth of all atoms in a residue
        ca_depth: depth of the alpha carbon atom

    Depths are in units Angstroms. 1A = 10^-10 m = 1nm

    Args:
        pdb_file: Path to PDB file
        outfile: Optional name of output file (without extension)
        outdir: Optional output directory
        outext: Optional extension for the output file
        outext: Suffix appended to json results file
        force_rerun: Rerun MSMS even if results exist already

    Returns:
        Pandas DataFrame: ResidueDepth property_dict, reformatted

    """
    # Create the output file name
    outfile = ssbio.utils.outfile_maker(inname=pdb_file, outname=outfile, outdir=outdir, outext=outext)

    if ssbio.utils.force_rerun(flag=force_rerun, outfile=outfile):
        # Load the structure
        my_structure = StructureIO(pdb_file)
        model = my_structure.first_model
        df = get_msms_df(model, pdb_file, outfile=outfile, outdir=outdir, outext=outext, force_rerun=force_rerun)
    else:
        log.debug('{}: already ran MSMS and force_rerun={}, loading results'.format(outfile, force_rerun))
        df = pd.read_csv(outfile, index_col=0)

    return df


if __name__ == '__main__':
    p = argparse.ArgumentParser(description='Run MSMS to calculate residue depth on a file, files, or a directory. Save a JSON file of results per structure.')
    p.add_argument('infile', help='PDB file, files, or folder', type=str, nargs='+')
    p.add_argument('--summary', action='store_true', help='Save a summary DataFrame of results.')

    args = p.parse_args()
    infiles = ssbio.utils.input_list_parser(args.infile)

    msms_errors = []

    for f in tqdm(infiles):
        msms_stuff = get_msms_df_on_file(f)
        if msms_stuff.empty:
            msms_errors.append(f)

    if args.summary:
        pass
        # TODO: what to save as a summary? Average residue depth of entire protein?
        log.info('Saved DF at: {}'.format('{}-df_msms_summary.csv'.format(ssbio.utils.todays_short_date())))

    if msms_errors:
        log.warning('Errors with: {}'.format(msms_errors))


import warnings

import numpy
from Bio.PDB.MMCIFParser import MMCIFParser
from Bio.PDB.PDBExceptions import PDBConstructionException
from Bio.PDB.PDBExceptions import PDBConstructionWarning

from ssbio.biopython.bp_mmcif2dict import MMCIF2DictFix


class MMCIFParserFix(MMCIFParser):
    """Fixes for MMCIFParser according to biopython#481 and biopython#523. Methods override parent.
    """

    def __init__(self, QUIET=True):
        super(MMCIFParserFix, self).__init__(QUIET=QUIET)

    def get_structure(self, structure_id, filename):
        with warnings.catch_warnings():
            if self.QUIET:
                warnings.filterwarnings("ignore", category=PDBConstructionWarning)
        self._mmcif_dict = MMCIF2DictFix(filename)
        self._build_structure(structure_id)
        return self._structure_builder.get_structure()

    def _build_structure(self, structure_id):
        mmcif_dict = self._mmcif_dict
        atom_id_list = mmcif_dict["_atom_site.label_atom_id"]
        residue_id_list = mmcif_dict["_atom_site.label_comp_id"]
        try:
            element_list = mmcif_dict["_atom_site.type_symbol"]
        except KeyError:
            element_list = None
        seq_id_list = mmcif_dict["_atom_site.label_seq_id"]
        auth_seq_id_list = mmcif_dict["_atom_site.auth_seq_id"]
        chain_id_list = mmcif_dict["_atom_site.auth_asym_id"]
        x_list = [float(x) for x in mmcif_dict["_atom_site.Cartn_x"]]
        y_list = [float(x) for x in mmcif_dict["_atom_site.Cartn_y"]]
        z_list = [float(x) for x in mmcif_dict["_atom_site.Cartn_z"]]
        alt_list = mmcif_dict["_atom_site.label_alt_id"]
        icode_list = mmcif_dict["_atom_site.pdbx_PDB_ins_code"]
        b_factor_list = mmcif_dict["_atom_site.B_iso_or_equiv"]
        occupancy_list = mmcif_dict["_atom_site.occupancy"]
        fieldname_list = mmcif_dict["_atom_site.group_PDB"]
        try:
            serial_list = [int(n) for n in mmcif_dict["_atom_site.pdbx_PDB_model_num"]]
        except KeyError:
            # No model number column
            serial_list = None
        except ValueError:
            # Invalid model number (malformed file)
            raise PDBConstructionException("Invalid model number")
        try:
            aniso_u11 = mmcif_dict["_atom_site.aniso_U[1][1]"]
            aniso_u12 = mmcif_dict["_atom_site.aniso_U[1][2]"]
            aniso_u13 = mmcif_dict["_atom_site.aniso_U[1][3]"]
            aniso_u22 = mmcif_dict["_atom_site.aniso_U[2][2]"]
            aniso_u23 = mmcif_dict["_atom_site.aniso_U[2][3]"]
            aniso_u33 = mmcif_dict["_atom_site.aniso_U[3][3]"]
            aniso_flag = 1
        except KeyError:
            # no anisotropic B factors
            aniso_flag = 0
        # if auth_seq_id is present, we use this.
        # Otherwise label_seq_id is used.
        if "_atom_site.auth_seq_id" in mmcif_dict:
            seq_id_list = mmcif_dict["_atom_site.auth_seq_id"]
        else:
            seq_id_list = mmcif_dict["_atom_site.label_seq_id"]
        # Now loop over atoms and build the structure
        structure_builder = self._structure_builder
        structure_builder.init_structure(structure_id)
        structure_builder.init_seg(" ")
        # Historically, Biopython PDB parser uses model_id to mean array index
        # so serial_id means the Model ID specified in the file
        current_model_id = -1
        current_serial_id = None
        current_chain_id = None
        current_residue_id = None
        current_icode = None
        for i in range(0, len(atom_id_list)):

            # set the line_counter for 'ATOM' lines only and not
            # as a global line counter found in the PDBParser()
            # this number should match the '_atom_site.id' index in the MMCIF
            structure_builder.set_line_counter(i)

            x = x_list[i]
            y = y_list[i]
            z = z_list[i]
            resname = residue_id_list[i]
            chainid = chain_id_list[i]
            altloc = alt_list[i]
            if altloc == ".":
                altloc = " "
            try:
                resseq = int(seq_id_list[i])
            except ValueError:
                resseq = seq_id_list[i]
            icode = icode_list[i]
            if icode == "?":
                icode = " "
            name = atom_id_list[i]
            # occupancy & B factor
            try:
                tempfactor = float(b_factor_list[i])
            except ValueError:
                raise PDBConstructionException("Invalid or missing B factor")
            try:
                occupancy = float(occupancy_list[i])
            except ValueError:
                raise PDBConstructionException("Invalid or missing occupancy")
            fieldname = fieldname_list[i]
            if fieldname == "HETATM":
                hetatm_flag = "H"
            else:
                hetatm_flag = " "
            if serial_list is not None:
                # model column exists; use it
                serial_id = serial_list[i]
                if current_serial_id != serial_id:
                    # if serial changes, update it and start new model
                    current_serial_id = serial_id
                    current_model_id += 1
                    current_chain_id = None
                    current_residue_id = None
                    current_icode = None
                    structure_builder.init_model(current_model_id, current_serial_id)
            else:
                # no explicit model column; initialize single model
                current_model_id = 0
                structure_builder.init_model(current_model_id)

            if current_chain_id != chainid:
                current_chain_id = chainid
                current_residue_id = None
                current_icode = None
                structure_builder.init_chain(current_chain_id)

            if current_residue_id != resseq:
                current_residue_id = resseq
                current_icode = None
                if current_icode != icode:
                    current_icode = icode
                structure_builder.init_residue(resname, hetatm_flag, resseq, icode)
            else:
                if current_icode != icode:
                    current_icode = icode
                    structure_builder.init_residue(resname, hetatm_flag, resseq, icode)

            coord = numpy.array((x, y, z), 'f')
            element = element_list[i] if element_list else None
            structure_builder.init_atom(name, coord, tempfactor, occupancy, altloc,
                                        name, element=element)
            if aniso_flag == 1:
                u = (aniso_u11[i], aniso_u12[i], aniso_u13[i],
                     aniso_u22[i], aniso_u23[i], aniso_u33[i])
                mapped_anisou = [float(x) for x in u]
                anisou_array = numpy.array(mapped_anisou, 'f')
                structure_builder.set_anisou(anisou_array)
        # Now try to set the cell
        try:
            a = float(mmcif_dict["_cell.length_a"])
            b = float(mmcif_dict["_cell.length_b"])
            c = float(mmcif_dict["_cell.length_c"])
            alpha = float(mmcif_dict["_cell.angle_alpha"])
            beta = float(mmcif_dict["_cell.angle_beta"])
            gamma = float(mmcif_dict["_cell.angle_gamma"])
            cell = numpy.array((a, b, c, alpha, beta, gamma), 'f')
            spacegroup = mmcif_dict["_symmetry.space_group_name_H-M"]
            spacegroup = spacegroup[1:-1]  # get rid of quotes!!
            if spacegroup is None:
                raise Exception
            structure_builder.set_symmetry(spacegroup, cell)
        except Exception:
            pass  # no cell found, so just ignore
